from lafhterlearn.evaluation import evaluate
from lafhterlearn.training import print_metrics
from lafhterlearn.environment import Environment
from lafhterlearn.session import SessionDirectoryLayout

from .base import Command


class PretrainCommand(Command):
    name = 'pretrain'
    help = 'Start/resume pretraining the model on synthetic data'

    def configure_parser(self, parser):
        configure_parser(parser)

    def __call__(self, args):
        run(args)


def configure_parser(parser):
    parser.add_argument('session_dir', type=str, default='',
                        help='Location of the session directory')


def run(args):
    config = SessionDirectoryLayout(args.session_dir).load_config()
    env = Environment(config)
    training_loop = env.training_loop

    for epoch in training_loop:
        metrics = {}
        for task in env.eval_tasks:
            metrics.update(evaluate(task))

        print_metrics(metrics, epoch)
        env.history_saver.add_entry(epoch, metrics)
        env.save_checkpoint(epoch, metrics)

    print("Done")



from random import randrange, uniform
import math
import random
from dataclasses import dataclass
import torch

from torchvision import transforms
from torchvision.transforms import Compose, PILToTensor, ToPILImage
from .image_utils import fit_height


@dataclass
class WeakAugmentation:
    p_augment: float
    target_height: int
    fill: int
    rotation_degrees_range: tuple
    blur_size: int
    blur_sigma: tuple
    noise_sigma: int
    should_fit_height: bool

    def __call__(self, images):
        rotate = self._rotate_and_scale()
        blur = transforms.GaussianBlur(self.blur_size, sigma=self.blur_sigma)
        add_noise = gaussian_noise(sigma=self.noise_sigma)

        images = self._augment_images(images, rotate)
        images = self._augment_images(images, add_noise)
        images = self._augment_images(images, blur)

        return images

    def _augment_images(self, images, augment_func):
        return [augment_func(im) if self._should_augment() else im for im in images]

    def _should_augment(self):
        return random.random() < self.p_augment

    def _rotate_and_scale(self):
        rotate = transforms.RandomRotation(degrees=self.rotation_degrees_range,
                                           expand=True, fill=self.fill)

        def transform_func(image):
            return fit_height(rotate(image), target_height=self.target_height)

        func = transform_func if self.should_fit_height else rotate
        return func


class StrongAugmentation:
    brightness = transforms.ColorJitter(brightness=(0.05, 0.95))
    contrast = transforms.ColorJitter(contrast=(0.05, 0.95))
    equalize = transforms.RandomEqualize(1)
    rotate = transforms.RandomRotation((-30, 30), fill=255)

    degrees_range = (math.degrees(-0.3), math.degrees(0.3))
    shear_x = transforms.RandomAffine(0, shear=degrees_range, fill=255)
    shear_y = transforms.RandomAffine(0, shear=(0, 0) + degrees_range, fill=255)
    auto_contrast = transforms.RandomAutocontrast(1)
    translate_x = transforms.RandomAffine(0, translate=(0.3, 0), fill=255)
    translate_y = transforms.RandomAffine(0, translate=(0, 0.3), fill=255)

    transforms_per_image = 2

    def __call__(self, images):
        return [self.transform_image(im) for im in images]

    def transform_image(self, image):
        transformations = self.get_random_transformations(self.transforms_per_image)
        for transform_func in transformations:
            image = transform_func(image)
        return image

    def get_random_transformations(self, n):
        return [self.random_transformation() for _ in range(n)]

    def random_transformation(self):
        all_transforms = [self.auto_contrast, self.brightness, self.contrast,
                          self.equalize, identity, posterize, self.rotate,
                          adjust_sharpness, self.shear_x, self.shear_y,
                          solarize, self.translate_x, self.translate_y]
        idx = randrange(0, len(all_transforms))
        return all_transforms[idx]


def identity(image): return image


def posterize(image):
    bits = randrange(4, 8 + 1)
    return transforms.RandomPosterize(bits, p=1)(image)


def adjust_sharpness(image):
    factor = uniform(0.05, 0.95)
    return transforms.RandomAdjustSharpness(factor, p=1)(image)


def solarize(image):
    threshold = int(round(uniform(0, 1) * 255))
    return transforms.RandomSolarize(threshold, p=1)(image)


def gaussian_noise(sigma):
    def add_noise(tensor):
        noisy = tensor + sigma * torch.randn_like(tensor.to(torch.float32))
        noisy = torch.clamp(noisy, 0, 255)
        return noisy.to(tensor.dtype)

    return Compose([PILToTensor(), add_noise, ToPILImage()])



from dataclasses import dataclass

import torch

from PIL import Image

from lafhterlearn.training import TrainingLoop
from lafhterlearn.evaluation import evaluate
from lafhterlearn.training import print_metrics
from lafhterlearn.environment import TuningEnvironment
from lafhterlearn.formatters import show_progress_bar
from lafhterlearn.decoding import decode_and_score
from lafhterlearn.datasets import PseudoLabeledDataset
from lafhterlearn.session import SessionDirectoryLayout
from .base import Command


class FinetuneCommand(Command):
    name = 'finetune'
    help = 'Finetune pretrained neural net on real unlabeled handwriting images'

    def configure_parser(self, parser):
        configure_parser(parser)

    def __call__(self, args):
        run(args)


def configure_parser(parser):
    parser.add_argument('session_dir', type=str,
                        help='Location of the session directory')


def run(args):
    session_config = SessionDirectoryLayout(args.session_dir).load_config()
    fine_tune(session_config)


def predict_labels(data_loader, recognizer, tokenizer):
    recognizer.neural_pipeline.eval_mode()

    for data in show_progress_bar(data_loader, desc='Predicting pseudo labels: '):
        image_paths = data[0]
        images = [Image.open(path) for path in image_paths]
        y_hat = recognizer(images)
        transcripts, scores = decode_and_score(y_hat, tokenizer)

        yield PseudoLabeledBatch(image_paths=image_paths, transcripts=transcripts, scores=scores)


@dataclass
class PseudoLabeledBatch:
    image_paths: list
    transcripts: list
    scores: list

    def above_threshold(self, threshold):
        for path, transcript, score in zip(self.image_paths, self.transcripts, self.scores):
            if score > threshold:
                yield path, transcript


def train_on_pseudo_labels(env, epoch, pseudo_labeled):
    pseudo_labeled_ds = PseudoLabeledDataset(pseudo_labeled)
    trainer = env.get_trainer(pseudo_labeled_ds)
    training_loop = TrainingLoop(trainer, env.metric_fns, epochs=epoch + 1, starting_epoch=epoch)
    next(iter(training_loop))


def compute_metrics(env):
    metrics = {}
    for task in env.tasks:
        metrics.update(evaluate(task))
    return metrics


def fine_tune(config):
    env = TuningEnvironment(config)

    starting_epoch = env.get_trained_epochs() + 1
    for epoch in range(starting_epoch, env.tuning_epochs):
        pseudo_labeled = []
        with torch.no_grad():
            for pseudo_labeled_batch in predict_labels(env.unlabeled_loader,
                                                       env.recognizer,
                                                       env.tokenizer):
                paths_with_transcripts = pseudo_labeled_batch.above_threshold(env.threshold)
                pseudo_labeled.extend(paths_with_transcripts)

        train_on_pseudo_labels(env, epoch, pseudo_labeled)

        metrics = compute_metrics(env)
        print_metrics(metrics, epoch)
        env.history_saver.add_entry(epoch, metrics)
        env.save_checkpoint(epoch, metrics)



import torch
from torchvision.transforms import Resize, Pad, ToTensor


def clip_height(image, max_value):
    """Ensure image has at most max_value height"""
    if image.height > max_value:
        image = fit_height(image, max_value)
    return image


def clip_all_heights(images, max_height):
    """Ensure every image has at most max_value height"""
    return [clip_height(im, max_height) for im in images]


def fit_height(image, target_height):
    """Resize image to have target_height height and keep its aspect ratio"""
    w = image.width
    h = image.height

    scaler = target_height / h

    target_width = int(round(scaler * w))

    resizer = Resize((target_height, target_width))
    return resizer(image)


def resize_all(images, height):
    """Resize all images to specified height whilst retaining the aspect ratio"""
    return [fit_height(im, height) for im in images]


def pad_images(images, pad_strategy=None, max_width=None, max_height=None,
               fill=255, extra_pad=0):
    pad_strategy = pad_strategy or one_sided_padding

    max_height = max_height or compute_max_height(images)

    max_batch_width = compute_max_width(images)

    if not (max_width and max_width >= max_batch_width):
        max_width = max_batch_width
    max_width = nearest_divisible_by(max_width)

    # give extra padding to each image signaling the end of text
    max_width = max_width + extra_pad

    padded = []
    for im in images:
        padding_top, padding_bottom = pad_strategy(max_height, im.height)
        padding_left, padding_right = pad_strategy(max_width, im.width)

        pad = Pad((padding_left, padding_top, padding_right, padding_bottom),
                  fill=fill)
        padded.append(pad(im))

    return padded


def compute_max_height(images):
    return max([im.height for im in images])


def compute_max_width(images):
    return max([im.width for im in images])


def nearest_divisible_by(value, factor=32):
    """Calculate the smallest integer larger than or equal to the value divisible by factor"""
    if value % factor != 0:
        value = (value // factor + 1) * factor
    return value


def equal_padding(max_length, length):
    len_diff = max_length - length
    padding1 = len_diff // 2
    padding2 = len_diff - padding1
    return padding1, padding2


def one_sided_padding(max_length, length):
    len_diff = max_length - length
    return 0, len_diff


def to_rgb(tensor):
    """Convert tensor with single channel to tensor with 3 channels.

    Simply repeat same channel 3 times

    :param tensor: Corresponding to the image tensor of shape (1, height, width)
    :return: Tensor of shape (3, height, width)
    """
    return tensor.repeat(3, 1, 1)


def to_tensors(images):
    """Convert all images to tensors with 1 channel"""
    to_tensor = ToTensor()
    return [to_tensor(im) for im in images]


def to_rgb_tensors(tensors):
    """Convert all images to tensors consisting of 3 identical channels"""
    return [to_rgb(t) for t in tensors]


def make_batch(images):
    """Convert a list of images to tensor of shape (num_images, 1, height, width).

    All images are expected to have the same size.
    """
    tensors = to_tensors(images)
    return torch.stack(tensors)


def make_rgb_batch(images):
    """Convert a list of images to tensor of shape (num_images, 3, height, width).

    All images are expected to have the same size.
    """
    tensors = to_tensors(images)
    tensors = to_rgb_tensors(tensors)
    return torch.stack(tensors)



from functools import partial

from .augmentation import WeakAugmentation
from .image_utils import clip_all_heights, resize_all, pad_images, make_rgb_batch


class ImagePipeline:
    def __init__(self, transform_functions):
        self.transform_functions = transform_functions

    def __call__(self, images):
        for func in self.transform_functions:
            images = func(images)
        return images


def compose_transforms(max_width, max_height, resize=False):
    def clip(images):
        return clip_all_heights(images, max_height=max_height)

    def pad(images):
        return pad_images(images, max_width=max_width, max_height=max_height, extra_pad=32)

    if resize and max_height:
        initial_transform = partial(resize_all, height=max_height)
    else:
        initial_transform = clip

    return [initial_transform] + [pad, make_rgb_batch]


def make_pretraining_pipeline(augmentation_options, max_width, max_height, resize=False):
    no_augment_transforms = compose_transforms(max_width, max_height, resize)

    augment = WeakAugmentation(**augmentation_options)
    all_transforms = [augment] + no_augment_transforms

    return ImagePipeline(all_transforms)


def make_validation_pipeline(max_width, max_height, resize=False):
    return ImagePipeline(compose_transforms(max_width, max_height, resize))



import argparse

from .base import commands
from . import (data, evaluate, finetune, extract_fonts, pretrain,
               make_session, transcribe, visualize_fonts, word_distr, learn_ngrams)


def parse_all_args():
    root_parser = argparse.ArgumentParser(
        description='A set of tools for training a neural net for handwriting recognition'
    )
    all_subparsers = root_parser.add_subparsers(dest='command')

    for name, command in commands.items():
        command.setup(all_subparsers)

    return root_parser.parse_args()


def execute_command():
    args = parse_all_args()
    command = commands[args.command]
    command(args)



import os
from PIL import Image
from torch.utils.data import Dataset
from .data_generator import SimpleRandomWordGenerator


class SyntheticOnlineDataset(Dataset):
    def __init__(self, fonts_dir, size, word_sampler,
                 bg_range=(255, 255),
                 color_range=(0, 100),
                 stroke_fill_range=(0, 100),
                 font_size_range=(50, 100),
                 rotation_range=(0, 0),
                 spaces_range=(0, 2)):
        super().__init__()
        self.size = size
        self.fonts_dir = fonts_dir

        simple_generator = SimpleRandomWordGenerator(word_sampler, self.fonts_dir,
                                                     bg_range=bg_range,
                                                     color_range=color_range,
                                                     stroke_fill_range=stroke_fill_range,
                                                     font_size_range=font_size_range,
                                                     rotation_range=rotation_range,
                                                     spaces_range=spaces_range)
        self.iterator = iter(simple_generator)

    def __getitem__(self, idx):
        if 0 <= idx < len(self):
            return self.generate_example()
        else:
            raise IndexError()

    def generate_example(self):
        im, word = next(self.iterator)
        return im, word

    def __len__(self):
        return self.size


class SyntheticOnlineDatasetCached(SyntheticOnlineDataset):
    def __init__(self, fonts_dir, size, word_sampler, bg_range=(255, 255),
                 color_range=(0, 100), stroke_fill_range=(0, 100),
                 font_size_range=(50, 100), rotation_range=(0, 0),
                 spaces_range=(0, 2)):
        super().__init__(fonts_dir, size, word_sampler,
                         bg_range=bg_range,
                         color_range=color_range,
                         stroke_fill_range=stroke_fill_range,
                         font_size_range=font_size_range,
                         rotation_range=rotation_range,
                         spaces_range=spaces_range)

        self.cache = {}

    def __getitem__(self, idx):
        if idx not in self.cache:
            self.cache[idx] = super().__getitem__(idx)

        return self.cache[idx]


class RealLabeledDataset(Dataset):
    def __init__(self, ds_dir):
        self.examples = self.load_data(ds_dir)

    def load_data(self, ds_dir):
        transcripts_path = os.path.join(ds_dir, "transcripts.txt")
        transcripts_exist = os.path.isfile(transcripts_path)
        loader = PreparedDatasetLoader() if transcripts_exist else ImageFolderLoader()
        return loader(ds_dir)

    def __getitem__(self, idx):
        image_path, transcript = self.examples[idx]
        image = Image.open(image_path)
        return image, transcript

    def __len__(self):
        return len(self.examples)


class ImageFolderLoader:
    """Assumes that dataset is simply set of images in a folder.
    Transcript of each image is contained in its file name.
    For example, an image with a file name "apple_84" will have
    its label "apple".
    """
    def __call__(self, ds_dir):
        examples = []
        for f_name in os.listdir(ds_dir):
            if f_name.endswith('.png') or f_name.endswith('.jpg'):
                path = os.path.join(ds_dir, f_name)
                s, _ = os.path.splitext(f_name)
                transcript = s.split('_')[0]
                examples.append((path, transcript))
        return examples


class PreparedDatasetLoader:
    """Assumes that dataset is prepared by the toolkit"""
    def __call__(self, ds_dir):
        transcripts_path = os.path.join(ds_dir, "transcripts.txt")
        with open(transcripts_path) as f:
            content = f.read()
            transcripts = content.split("\n")

        def sort_key(file_name):
            name, _ = os.path.splitext(file_name)
            return int(name)

        file_names = [f_name for f_name in os.listdir(ds_dir)
                      if not f_name.endswith('.txt')]

        file_names.sort(key=sort_key)
        image_paths = [os.path.join(ds_dir, f_name) for f_name in file_names]
        return list(zip(image_paths, transcripts))


class RealUnlabeledDataset(Dataset):
    def __init__(self, ds_dir):
        self.image_paths = [os.path.join(ds_dir, f_name) for f_name in os.listdir(ds_dir)]

    def __getitem__(self, idx):
        return self.image_paths[idx],

    def __len__(self):
        return len(self.image_paths)


class PseudoLabeledDataset(Dataset):
    def __init__(self, paths_with_transcripts):
        self.paths_with_transcripts = paths_with_transcripts

    def __getitem__(self, idx):
        path, transcript = self.paths_with_transcripts[idx]
        image = Image.open(path)
        return image, transcript

    def __len__(self):
        return len(self.paths_with_transcripts)



import itertools
from unittest import TestCase
import h5py
import os
from lafhterlearn.metrics import Metric
from lafhterlearn.ngram_utils import ExpandableMatrix, RaggedMatrix, SparseMatrix
from lafhterlearn.ngrams import chunks, sorted_chunks, merge_chunks, bigsort, ChainSequence


class ChainSequenceTests(TestCase):
    def test_len(self):
        seq = ChainSequence()
        self.assertEqual(0, len(seq))

        seq = ChainSequence([])
        self.assertEqual(0, len(seq))

        seq = ChainSequence([], [], [])
        self.assertEqual(0, len(seq))

        seq = ChainSequence([1, 2, 3])
        self.assertEqual(3, len(seq))

        seq = ChainSequence([1, 2], [3, 4, 5])
        self.assertEqual(5, len(seq))

        seq = ChainSequence([1, 2, 3], [0], [], [0])
        self.assertEqual(5, len(seq))

    def test_indexing(self):
        seq = ChainSequence()
        self.assertRaises(IndexError, lambda: seq[0])

        seq = ChainSequence([])
        self.assertRaises(IndexError, lambda: seq[0])

        seq = ChainSequence([], [], [])
        self.assertRaises(IndexError, lambda: seq[0])

        seq = ChainSequence([30])
        self.assertEqual(30, seq[0])

        seq = ChainSequence([5, 3], [10, 20, 30])
        self.assertEqual(5, seq[0])
        self.assertEqual(3, seq[1])
        self.assertEqual(10, seq[2])
        self.assertEqual(20, seq[3])
        self.assertEqual(30, seq[4])

        seq = ChainSequence([2], [0], [12], [99], [33])
        self.assertEqual(2, seq[0])
        self.assertEqual(0, seq[1])
        self.assertEqual(12, seq[2])
        self.assertEqual(99, seq[3])
        self.assertEqual(33, seq[4])

        seq = ChainSequence([1, 2], [5], [12, 17, 20, 34], [99])
        self.assertEqual(1, seq[0])
        self.assertEqual(2, seq[1])
        self.assertEqual(5, seq[2])
        self.assertEqual(12, seq[3])
        self.assertEqual(17, seq[4])
        self.assertEqual(20, seq[5])
        self.assertEqual(34, seq[6])
        self.assertEqual(99, seq[7])


class ChunkingTests(TestCase):
    def test(self):
        self.assertRaises(ValueError, lambda: list(chunks([1, 2, 3], 0)))

        self.assertEqual([], list(chunks([], 5)))

        self.assertEqual([[1]], list(chunks([1], 1)))
        self.assertEqual([[1]], list(chunks([1], 2)))

        self.assertEqual([[1], [2]], list(chunks([1, 2], 1)))
        self.assertEqual([[1, 2]], list(chunks([1, 2], 2)))
        self.assertEqual([[1, 2]], list(chunks([1, 2], 3)))

        self.assertEqual([[1], [2], [3]], list(chunks([1, 2, 3], 1)))
        self.assertEqual([[1, 2], [3]], list(chunks([1, 2, 3], 2)))
        self.assertEqual([[1, 2, 3]], list(chunks([1, 2, 3], 3)))
        self.assertEqual([[1, 2, 3]], list(chunks([1, 2, 3], 4)))

        self.assertEqual([[1], [2], [3], [4]], list(chunks([1, 2, 3, 4], 1)))
        self.assertEqual([[1, 2], [3, 4]], list(chunks([1, 2, 3, 4], 2)))
        self.assertEqual([[1, 2, 3], [4]], list(chunks([1, 2, 3, 4], 3)))
        self.assertEqual([[1, 2, 3, 4]], list(chunks([1, 2, 3, 4], 4)))


class SortedChunksTests(TestCase):
    def test(self):
        self.assertEqual([[1, 2, 3], [2, 3, 4]], list(sorted_chunks([3, 2, 1, 2, 4, 3], 3)))
        self.assertEqual([[1, 2, 3], [2, 4]], list(sorted_chunks([3, 2, 1, 2, 4], 3)))
        self.assertEqual([[1, 2, 2, 3], [4]], list(sorted_chunks([3, 2, 1, 2, 4], 4)))
        self.assertEqual([[3], [2], [1], [2]], list(sorted_chunks([3, 2, 1, 2], 1)))


class MergeChunksTests(TestCase):
    def test(self):
        self.assertEqual([1, 3, 4, 4, 5, 5], list(merge_chunks([1, 3, 4, 4, 5, 5])))

        self.assertEqual([], list(merge_chunks([])))
        self.assertEqual([], list(merge_chunks([], [])))
        self.assertEqual([], list(merge_chunks([], [], [], [], [], [])))

        self.assertEqual([1, 2], list(merge_chunks([1, 2], [])))
        self.assertEqual([-2, 0, 1, 2], list(merge_chunks([], [], [1, 2], [], [-2, 0])))

        self.assertEqual([1, 2, 3, 4], list(merge_chunks([1, 2], [3, 4])))
        self.assertEqual([1, 2, 3, 4], list(merge_chunks([3, 4], [1, 2])))

        self.assertEqual([0, 1, 1, 2, 3, 4, 4, 6], list(merge_chunks([1, 3, 4], [0, 1, 2, 4, 6])))

        merged = merge_chunks([1, 2, 4, 7, 8], [3, 5, 6, 9])
        self.assertEqual([1, 2, 3, 4, 5, 6, 7, 8, 9], list(merged))

        merged = merge_chunks([5, 9], [3, 3, 4], [-2, 0, 4, 5], [1])
        self.assertEqual([-2, 0, 1, 3, 3, 4, 4, 5, 5, 9], list(merged))

        merged = merge_chunks([5], [4], [3], [2], [1])
        self.assertEqual([1, 2, 3, 4, 5], list(merged))

        merged = merge_chunks([1], [2], [3], [4], [5])
        self.assertEqual([1, 2, 3, 4, 5], list(merged))

        merged = merge_chunks([1, 2], [3, 4, 5])
        self.assertEqual([1, 2, 3, 4, 5], list(merged))

        p = next(itertools.permutations(range(100)))

        merged = merge_chunks(*list(sorted_chunks(p, size=23)))
        self.assertEqual(list(range(100)), list(merged))


class BigsortTests(TestCase):
    def setUp(self) -> None:
        if os.path.isfile("chunks.h5"):
            os.remove("chunks.h5")

    def tearDown(self) -> None:
        if os.path.isfile("chunks.h5"):
            os.remove("chunks.h5")

    def test_sort_integers(self):
        gen = bigsort([2, 4, 3, 5, 6, 7, 1, 0, 8, 9], chunk_size=3)
        self.assertEqual(list(range(10)), list(gen))

    def test_sort_numeric_tuples(self):
        gen = bigsort([(3, 5), (1, 3), (2, 4), (1, 2), (3, 1)], chunk_size=3)
        self.assertEqual([(1, 2), (1, 3), (2, 4), (3, 1), (3, 5)], list(gen))


class ExpandableMatrixTests(TestCase):
    file_name = "testfile.h5"

    def setUp(self) -> None:
        f = h5py.File(self.file_name, mode='w')
        f.close()

    def tearDown(self) -> None:
        os.remove(self.file_name)

    def test_initial(self):
        with h5py.File(self.file_name, mode="r+") as f:
            matrix = ExpandableMatrix(f, "ds", 2)
            self.assertEqual(0, len(matrix))

    def test_read_from_empty(self):
        with h5py.File(self.file_name, mode="r+") as f:
            matrix = ExpandableMatrix(f, "ds", 2)
            self.assertRaises(IndexError, lambda: matrix[0])

    def test_add_single_row(self):
        with h5py.File(self.file_name, mode="r+") as f:
            matrix = ExpandableMatrix(f, "ds", 3)

            self.assertRaises(ValueError, lambda: matrix.append_row([1]))
            self.assertRaises(ValueError, lambda: matrix.append_row([1, 10]))
            self.assertRaises(ValueError, lambda: matrix.append_row([1, 10, 100, 10]))

            matrix.append_row([1, 2, 4])
            self.assertEqual(1, len(matrix))
            self.assertEqual([1, 2, 4], list(matrix[0]))

    def test_two_rows(self):
        with h5py.File(self.file_name, mode="r+") as f:
            matrix = ExpandableMatrix(f, "ds", 2)
            matrix.append_row([5, 10])
            matrix.append_row([20, 30])

            self.assertEqual(2, len(matrix))
            self.assertEqual([5, 10], list(matrix[0]))
            self.assertEqual([20, 30], list(matrix[1]))


class H5MatrixTests(TestCase):
    file_name = "testfile.h5"

    def setUp(self) -> None:
        f = h5py.File(self.file_name, mode='w')
        f.close()

    def tearDown(self) -> None:
        os.remove(self.file_name)

    def test_initial(self):
        with h5py.File(self.file_name, mode="r+") as f:
            matrix = RaggedMatrix(f, "ds")
            self.assertEqual(0, len(matrix))

    def test_read_from_empty(self):
        with h5py.File(self.file_name, mode="r+") as f:
            matrix = RaggedMatrix(f, "ds")
            self.assertRaises(IndexError, lambda: matrix[0])
            self.assertRaises(IndexError, lambda: matrix[20])
            self.assertRaises(IndexError, lambda: matrix[0, 0])

    def test_add_single_row(self):
        with h5py.File(self.file_name, mode="r+") as f:
            matrix = RaggedMatrix(f, "ds")
            matrix.append_row([1, 2, 4])
            self.assertEqual(1, len(matrix))
            self.assertEqual([1, 2, 4], list(matrix[0]))

    def test_add_empty_row(self):
        with h5py.File(self.file_name, mode="r+") as f:
            matrix = RaggedMatrix(f, "ds")
            matrix.append_row([])
            self.assertEqual(1, len(matrix))
            self.assertEqual([], list(matrix[0]))

    def test_add_few_rows_of_different_length(self):
        with h5py.File(self.file_name, mode="r+") as f:
            matrix = RaggedMatrix(f, "ds")
            matrix.append_row([1, 2, 4])
            matrix.append_row([5, 6])
            matrix.append_row([9, 2, 0, 5])

            self.assertEqual(3, len(matrix))
            self.assertEqual([1, 2, 4], list(matrix[0]))
            self.assertEqual([5, 6], list(matrix[1]))
            self.assertEqual([9, 2, 0, 5], list(matrix[2]))

    def test_update_row(self):
        with h5py.File(self.file_name, mode="r+") as f:
            matrix = RaggedMatrix(f, "ds")
            matrix.append_row([1, 2, 4])
            matrix.append_row([5, 6])

            matrix.update_row(1, [3, 2, 5, 10])
            self.assertEqual(2, len(matrix))
            self.assertEqual([1, 2, 4], list(matrix[0]))
            self.assertEqual([3, 2, 5, 10], list(matrix[1]))

    def test_persistence(self):
        with h5py.File(self.file_name, mode="r+") as f:
            matrix = RaggedMatrix(f, "ds")
            matrix.append_row([1, 2, 4])

            del matrix
            new_matrix = RaggedMatrix(f, "ds")
            self.assertEqual(1, len(new_matrix))
            self.assertEqual([1, 2, 4], list(new_matrix[0]))


class SparseMatrixTests(TestCase):
    file_name = "test_file_sparse.h5"

    def setUp(self) -> None:
        f = h5py.File(self.file_name, mode='w')
        f.close()

    def tearDown(self) -> None:
        os.remove(self.file_name)

    def test_initial(self):
        with h5py.File(self.file_name, mode='w') as f:
            matrix = SparseMatrix(f, "group", 4)
            self.assertEqual(len(matrix), 0)

    def test_append_row(self):
        with h5py.File(self.file_name, mode='w') as f:
            matrix = SparseMatrix(f, "group", 5)
            matrix.add_empty_row()
            self.assertEqual(len(matrix), 1)
            self.assertEqual([0, 0, 0, 0, 0], list(matrix[0]))

            matrix.bulk_update(0, [0, 2, 3], [9, 8, 7])
            self.assertEqual(len(matrix), 1)
            self.assertEqual([9, 0, 8, 7, 0], list(matrix[0]))

    def test_append_2_rows(self):
        with h5py.File(self.file_name, mode='w') as f:
            matrix = SparseMatrix(f, "group", 4)
            matrix.add_empty_row()
            matrix.add_empty_row()
            matrix.bulk_update(0, [1, 2], [5, 10])
            matrix.bulk_update(1, [0, 2], [5, 10])

            self.assertEqual([0, 5, 10, 0], list(matrix[0]))
            self.assertEqual([5, 0, 10, 0], list(matrix[1]))

    def test_row_updates(self):
        with h5py.File(self.file_name, mode='w') as f:
            matrix = SparseMatrix(f, "group", 4)
            matrix.add_empty_row()
            matrix.add_empty_row()
            matrix.add_empty_row()
            matrix.bulk_update(1, [1, 2], [5, 10])
            matrix.bulk_update(1, [0, 2, 3], [100, 20, 50])
            self.assertEqual([0, 0, 0, 0], list(matrix[0]))
            self.assertEqual([100, 5, 20, 50], list(matrix[1]))
            self.assertEqual([0, 0, 0, 0], list(matrix[2]))

    def test_cannot_use_wrong_indices(self):
        with h5py.File(self.file_name, mode='w') as f:
            matrix = SparseMatrix(f, "group", 4)
            matrix.add_empty_row()
            self.assertRaises(IndexError, lambda: matrix.bulk_update(3, [3], [3]))
            self.assertRaises(IndexError, lambda: matrix.bulk_update(0, [1, 2, 6], [9, 8, 7]))


class MetricTests(TestCase):
    def test_metric_without_arguments(self):
        def metric_fn():
            return 32

        metric = Metric("metric", metric_fn, [], None)
        self.assertEqual(32, metric())

    def test_metric_with_one_argument(self):
        def metric_fn(x): return x**2

        metric = Metric("metric", metric_fn, ["x"], None)
        self.assertEqual(9, metric(x=3, y=9))

    def test_metric_with_two_arguments(self):
        def metric_fn(arg1, arg2): return arg1 - arg2

        metric = Metric("metric", metric_fn, ["x", "y"], None)
        self.assertEqual(-6, metric(x=3, y=9))

    def test_metric_with_transform(self):
        def metric_fn(arg1, arg2): return arg1 - arg2

        def transform(x):
            return x + 1, x - 1

        metric = Metric("metric", metric_fn, ["x"], transform_fn=transform)
        self.assertEqual(2, metric(x=3))



import torch
from torch import nn

from .utils import prepare_targets, pad_sequences, truncate_sequences
from .tokenizer import CharacterTokenizer


class MaskedCrossEntropy:
    def __init__(self, reduction='mean', label_smoothing=0.0):
        """

        :param reduction: defines which reduction to apply to elementwise losses:
            if set to "mean", mean will be taken across all dimensions,
            if set to "sum", all individual losses across all dimensions will be summed
            if set to 'none', no reduction will be performed
        """
        self.loss_function = nn.CrossEntropyLoss(reduction='none', label_smoothing=label_smoothing)
        self.reduction = reduction

    def __call__(self, y_hat, ground_true, mask):
        losses = self.loss_function(self.swap_axes(y_hat), ground_true)

        masked_loss = losses[mask.mask]
        if self.reduction == 'mean':
            return masked_loss.mean()
        elif self.reduction == 'sum':
            return masked_loss.sum()
        else:
            return masked_loss

    def swap_axes(self, t):
        return t.transpose(1, 2)


class LossTargetTransform:
    """Transform function (callable) invoked by Metric instance wrapping loss function.

    It takes 2 arguments: predictions and transcripts.
    Prediction is PyTorch tensor of shape (batch_size, max_steps, num_classes)
    containing raw unnormalized probabilities of different classes.
    Transcripts is a list of corresponding transcripts for each prediction in a batch.
    Each transcript is a Python string (str).

    The callable converts all transcripts into token sequences, pads them and
    wraps in PyTorch LongTensor class. It also creates a mask which is
    another PyTorch 2d Tensor which specifies the original (unpadded) length
    of each token sequence.

    The callable keeps predictions tensor unchanged.

    It returns 3-tuple (prediction, targets, mask)
    """
    def __init__(self, charset):
        self.tokenizer = CharacterTokenizer(charset)

    def __call__(self, *args):
        y_hat, transcripts = args

        tokens = prepare_targets(transcripts, self.tokenizer)

        filler = tokens[0][-1]

        prediction_num_steps = y_hat.shape[1]

        max_transcript_len = max(len(token_seq) for token_seq in tokens)

        # ensures that each transcript is at most of prediction_num_steps length
        if max_transcript_len > prediction_num_steps:
            tokens = truncate_sequences(tokens, prediction_num_steps)

        seqs, mask = pad_sequences(tokens, filler)

        # truncate the tensor along steps dim to the length of the longest transcript
        max_steps = len(seqs[0])
        y_hat = y_hat[:, :max_steps]

        target = torch.LongTensor(seqs)
        return [y_hat] + [target] + [mask]



from PIL import Image

from lafhterlearn.environment import TuningEnvironment
from lafhterlearn.decoding import decode_output_batch
from lafhterlearn.session import SessionDirectoryLayout
from lafhterlearn.tokenizer import CharacterTokenizer
from .base import Command


class TranscribeCommand(Command):
    name = 'transcribe'
    help = 'Recognize a handwritten text on a given word image'

    def configure_parser(self, parser):
        parser.add_argument('session_dir', type=str,
                            help='Path to the location of session directory containing model weights')
        parser.add_argument('image_path', type=str, help='Path to the location of image containing word')

    def __call__(self, args):
        config = SessionDirectoryLayout(args.session_dir).load_config()
        env = TuningEnvironment(config)
        recognizer = env.recognizer

        tokenizer = CharacterTokenizer(config.charset)

        image = Image.open(args.image_path)
        label_distr = recognizer([image])
        texts = decode_output_batch(label_distr, tokenizer)
        print('Predicted text:', texts[0])



from lafhterlearn.metrics import Metric
from lafhterlearn.utils import instantiate_class


def create_metric(name, metric_fn, transform_fn):
    return Metric(
        name, metric_fn=metric_fn, metric_args=["y_hat", "y"], transform_fn=transform_fn
    )


def create_optimizer(model, optimizer_conf):
    optimizer_name = optimizer_conf['class']
    kwargs = optimizer_conf['kwargs']
    return instantiate_class(optimizer_name, model.parameters(), **kwargs)


def prepare_loss(loss_conf):
    loss_class = loss_conf["class"]
    loss_kwargs = loss_conf["kwargs"]
    loss_function = instantiate_class(loss_class, **loss_kwargs)

    loss_transform_conf = loss_conf["transform"]
    transform_class = loss_transform_conf["class"]
    transform_kwargs = loss_transform_conf["kwargs"]
    loss_transform = instantiate_class(transform_class, **transform_kwargs)

    return create_metric('loss', loss_function, loss_transform)


def prepare_metrics(metrics_conf):
    metric_fns = {}
    for name, spec in metrics_conf.items():
        metric_class = spec['class']
        metric_args = spec.get('args', [])
        metric_kwargs = spec.get('kwargs', {})
        metric_fn = instantiate_class(metric_class, *metric_args, **metric_kwargs)

        transform_conf = spec['transform']
        transform_class = transform_conf['class']
        transform_kwargs = transform_conf['kwargs']
        transform_fn = instantiate_class(transform_class, **transform_kwargs)
        metric_fns[name] = create_metric(name, metric_fn, transform_fn)

    return metric_fns



import itertools
import os
import h5py
from nltk.util import ngrams
import numpy as np
from .ngram_utils import build_vocab, get_word_stream
from .formatters import show_progress_bar


def build_ngram_model(get_corpus, output_file, n):
    """Count ngrams and create count tables"""

    vocab = list(build_vocab(get_word_stream(get_corpus, n)))
    word2int = {word: i for i, word in enumerate(vocab)}

    with h5py.File(output_file, mode='w') as f:
        f.attrs["vocab"] = vocab

        unigram_counts = build_unigram_counts(get_corpus, word2int)

        f.create_dataset(f'{1}-grams', data=unigram_counts)

        for order in range(2, n + 1):
            group = f.create_group(f'{order}-grams')
            group.attrs['order'] = order
            build_ngram_counts(get_corpus, group, word2int, n=order)


def build_unigram_counts(get_corpus, word2int):
    counts = np.zeros(len(word2int), dtype=int)

    for word in get_word_stream(get_corpus, 1):
        token = word2int.get(word, word2int['<UNK>'])
        counts[token] += 1
    return counts


def build_ngram_counts(get_corpus, h5group, word2int, n):
    tokens = (word2int.get(word, word2int['<UNK>'])
              for word in get_word_stream(get_corpus, n))

    ngrams_with_counts = sorted_ngram_counts(tokens, n)

    context_group = h5group.create_group('contexts')
    address_group = h5group.create_group('addresses')
    count_group = h5group.create_group('counts')

    gen_with_progress = show_progress_bar(data_to_write(ngrams_with_counts),
                                          desc='Progress: ')

    for chunk_number, chunk in enumerate(chunks(gen_with_progress, 1024 * 100)):
        contexts = []
        addresses = []
        counts = []
        for context, address, chunk_counts in chunk:
            contexts.append(context)
            addresses.append(address)
            for c in chunk_counts:
                counts.append(c)

        context_group.create_dataset(f'chunk_{chunk_number}', data=contexts)
        address_group.create_dataset(f'chunk_{chunk_number}', data=addresses)
        count_group.create_dataset(f'chunk_{chunk_number}', data=counts)


class CountTable:
    def __init__(self, h5group):
        self.h5group = h5group

        context_group = h5group['contexts']
        address_group = h5group['addresses']
        count_group = h5group['counts']

        self.contexts = load_2d_matrix(context_group)
        self.addresses = load_2d_matrix(address_group)
        self.counts = load_2d_matrix(count_group)

    def get_counts(self, context):
        import bisect
        context = tuple(context)
        idx = bisect.bisect_left(self.contexts, context)
        if self.contexts[idx] != context:
            raise ValueError(f'cannot find counts for context {context}')

        start, end = self.addresses[idx]

        res = [self.counts[i] for i in range(start, end)]
        #return self.counts[start:end, :]
        return res


class NgramModel:
    class ProbabilityDistribution:
        def __init__(self, counts_array):
            self.counts_array = counts_array
            self.rng = np.random.default_rng()

        def sample(self):
            counts = self.counts_array
            pmf = counts / sum(counts)

            indices = list(range(len(counts)))
            return self.rng.choice(indices, p=pmf)

    def __init__(self, count_tables, vocab):
        self.count_tables = count_tables
        self.vocab = vocab
        self.word2id = {word: i for i, word in enumerate(self.vocab)}

    def tokenize(self, word):
        if word not in self.word2id:
            word = '<UNK>'
        return self.word2id[word]

    def lookup_word(self, word_id):
        try:
            return self.vocab[word_id]
        except IndexError:
            return '<UNK>'

    def p_next(self, context):
        order = len(context) + 1
        from .ngram_utils import SparseArray
        for i in range(order, 1, -1):
            count_table = self.count_tables[order]
            try:
                counts = count_table.get_counts(context)
                assert len(counts) > 0
                sparse_array = SparseArray(len(self.vocab))
                for idx, value in counts:
                    sparse_array[idx] = value

                a = np.array(list(sparse_array), dtype=int)
                return self.ProbabilityDistribution(a)
            except ValueError:
                context = context[1:]

        if not context:
            unigram_counts = self.count_tables[1]
            return self.ProbabilityDistribution(unigram_counts)
        raise Exception('can not build probability distribution')

    def generate(self, num_words):
        start_token = self.tokenize('<s>')
        max_order = len(self.count_tables) - 1
        context = tuple(start_token for _ in range(max_order))

        output = []
        for i in range(num_words):
            p = self.p_next(context)
            word_idx = p.sample()
            word = self.lookup_word(word_idx)
            context = context[1:] + (word_idx, )

            output.append(word)

        return output

    @classmethod
    def from_h5file(cls, f):
        unigram_counts = f["1-grams"]

        def get_order(name):
            return int(name.split('-')[0])

        names = sorted([group for group in f.keys() if group != '1-grams'],
                       key=get_order)

        count_tables = {
            1: unigram_counts
        }
        for name in names:
            order = get_order(name)
            group = f[name]
            count_tables[order] = CountTable(group)

        vocab = f.attrs['vocab']
        return cls(count_tables, vocab)


def load_2d_matrix(h5group):
    def parse_chunk_number(name):
        _, number = name.split('_')
        return int(number)

    seqs = []
    for k in sorted(h5group.keys(), key=parse_chunk_number):
        seqs.append(h5group[k])

    return ChainSequence(*seqs)


class ChainSequence:
    """Read-only super sequence composed of individual sequences"""
    def __init__(self, *seqs):
        seqs = [s for s in seqs if s]
        self.seqs = seqs

        lengths = [len(s) for s in seqs]
        cum_lengths = list(itertools.accumulate(lengths, initial=0))
        self.intervals = list(zip(cum_lengths[:-1], cum_lengths[1:]))

    def __getitem__(self, idx):
        if not (0 <= idx < len(self)):
            raise IndexError()

        for (a, b), seq in zip(self.intervals, self.seqs):
            if a <= idx < b:
                offset = idx - a
                return normalize(seq[offset])

        raise IndexError()

    def __len__(self):
        return sum(len(s) for s in self.seqs)


def data_to_write(ngrams_with_counts):
    pos = 0
    for context, sparse_array in group_by_contexts(ngrams_with_counts):
        start = pos
        end = pos + len(sparse_array)
        address = (start, end)
        pos += len(sparse_array)
        yield context, address, sparse_array


def group_by_contexts(ngrams_with_counts):
    groups = itertools.groupby(ngrams_with_counts, key=lambda t: t[0][:-1])

    for context, grouper in groups:
        counts = []
        for ngram, count in grouper:
            token = ngram[-1]
            counts.append((token, count))
        yield context, counts


def sorted_ngram_counts(tokens, n):
    sorted_ngrams = bigsort(ngrams(tokens, n))

    if os.path.isfile("chunks.h5"):
        os.remove("chunks.h5")

    for ngram, grouper in itertools.groupby(sorted_ngrams):
        yield ngram, len(list(grouper))


def bigsort(iterable, chunk_size=1024 * 100, temp_file="chunks.h5"):
    """Sort large collection of data that may not fit in memory.
    Generates and yields items in ascending order.
    """

    with h5py.File(temp_file, "a") as f:
        for i, chunk in enumerate(sorted_chunks(iterable, chunk_size)):
            f.create_dataset(f'chunk_{i}', data=list(chunk))

        datasets = [serializable_ds(f[ds_name]) for ds_name in f.keys()]
        yield from merge_chunks(*datasets)


def serializable_ds(ds):
    for item in ds:
        yield normalize(item)


def normalize(item):
    if isinstance(item, np.int64):
        item = int(item)
    elif isinstance(item, int):
        item = int(item)
    else:
        item = tuple(item)

    return item


def chunks(iterable, size):
    if size <= 0:
        raise ValueError(f'chunk size has to be positive integer, got {size}')

    s = []
    for item in iterable:
        s.append(item)

        if len(s) == size:
            yield s
            s = []

    if s:
        yield s


def sorted_chunks(iterable, size):
    for chunk in chunks(iterable, size):
        yield sorted(chunk)


def merge_chunks(*iterables):
    iterators = []
    items = []

    for it in iterables:
        iterator = iter(it)

        try:
            item = next(iterator)
            iterators.append(iterator)
            items.append(item)
        except StopIteration:
            pass

    while items:
        smallest = min(items)
        yield smallest

        iterator_index = items.index(smallest)
        try:
            items[iterator_index] = next(iterators[iterator_index])
        except StopIteration:
            iterators.pop(iterator_index)
            items.pop(iterator_index)



class Formatter:
    def __init__(self):
        self.progress_bar = ProgressBar()

    def format_epoch(self, epoch):
        return f'Epoch {epoch:5}'

    def format_metrics(self, metrics):
        metric_strings = [f'{name} {value:6.4f}' for name, value in metrics.items()]
        s = ', '.join(metric_strings)
        return f'[{s}]'

    def __call__(self, epoch, iteration, num_iterations, running_metrics):
        computed_metrics = {name: avg.value for name, avg in running_metrics.items()}

        metrics_str = self.format_metrics(computed_metrics)

        progress = self.progress_bar.updated(iteration, num_iterations)
        epoch_str = self.format_epoch(epoch)
        return f'{epoch_str} {metrics_str} {progress} {iteration} / {num_iterations}'


class ProgressBar:
    def updated(self, step, num_steps, fill='=', empty='.', cols=20):
        filled_chars = int(step / num_steps * cols)
        remaining_chars = cols - filled_chars
        return fill * filled_chars + '>' + empty * remaining_chars


def show_progress_bar(iterable, desc='', num_iters=None, cols=50):
    whitespaces = ' ' * 150
    print(f'\r{whitespaces}', end='')
    progress_bar = ProgressBar()

    if not num_iters and hasattr(iterable, '__len__'):
        num_iters = len(iterable)

    for i, data in enumerate(iterable):
        step_number = i + 1
        if num_iters:
            progress = progress_bar.updated(step_number, num_iters, cols=cols)
            msg = f'\r{desc}{progress} {step_number}/{num_iters}'
        else:
            max_val = 10000
            progress = progress_bar.updated(step_number % max_val, max_val, cols=cols)
            msg = f'\r{desc}{progress} Iterations {step_number}'
        print(msg, end='')
        yield data



import csv
import os

import nltk
from nltk.probability import FreqDist

from .base import Command


class CreateWordDistributionCommand(Command):
    name = 'word_distr'
    help = 'Prepare dictionary file or word distribution used to synthesize word images'

    def configure_parser(self, parser):
        configure_parser(parser)

    def __call__(self, args):
        run(args)


def configure_parser(parser):
    parser.add_argument('output_file', type=str, help='Location of output dictionary file')

    parser.add_argument('--dict-file', type=str, default='',
                        help='Location of a dictionary file (required if --corpora-dir is omitted)')

    parser.add_argument('--corpora-dir', type=str, default='',
                        help='Generate words from a folder of text files')

    parser.add_argument('--max-len', type=int, default=14,
                        help='Exclude words from the output file longer than max-len characters')
    parser.add_argument('--only-letters', default=False, action='store_true',
                        help='Filters out words containing digits and other non-letter characters')
    parser.add_argument('--with-freq', default=False, action='store_true',
                        help='Whether to calculate frequencies of each word')


def run(args):
    words_gen = load_corpora(args.corpora_dir) if args.corpora_dir else load_dict_file(args.dict_file)

    words_gen = filter_words(words_gen, args.only_letters, max_word_len=args.max_len)

    save_func = save_distr if args.with_freq else save_dict_file
    save_func(words_gen, args.output_file)


def load_dict_file(dict_file):
    """A generator yielding words from a dictionary text file"""
    with open(dict_file) as f:
        for word in f:
            if word.strip():
                yield word


def load_corpus(path):
    with open(path) as f:
        for line in f:
            line = line.strip()
            if line:
                yield from nltk.word_tokenize(line)


def load_corpora(corpora_dir):
    """A generator yielding words contained in text files located within corpora_dir folder"""
    for text_file in os.listdir(corpora_dir):
        path = os.path.join(corpora_dir, text_file)
        yield from load_corpus(path)


def filter_words(words, allow_only_letters=True, max_word_len=14):
    for w in words:
        too_long = len(w.strip()) > max_word_len
        has_non_letters = not w.isalpha()
        if too_long or (allow_only_letters and has_non_letters):
            continue
        yield w


def save_distr(words, destination):
    fdist = FreqDist()

    for i, word in enumerate(words):
        fdist[word] += 1

        if i % 100 == 0:
            print(f'\rCounted words: {i + 1}', end='')

    print('\nFrequency dictionary is complete')

    with open(destination, 'w', newline='') as csvfile:
        writer = csv.writer(csvfile, delimiter=',',
                            quotechar='|', quoting=csv.QUOTE_MINIMAL)

        for word in fdist.keys():
            p = fdist.freq(word)
            writer.writerow([word, p])


def save_dict_file(words, destination):
    vocab = set()
    for i, w in enumerate(words):
        vocab.add(w)
        if i % 100 == 0:
            print(f'\rBuilding vocab. Words processed: {i + 1}', end='')

    print('\nVocab is built')
    with open(destination, 'w') as f:
        file_content = '\n'.join(vocab)
        f.write(file_content)



commands = {}


def register_command(cls):
    instance = cls()
    commands[instance.name] = instance


class Meta(type):
    def __new__(meta, name, bases, class_dict):
        cls = type.__new__(meta, name, bases, class_dict)
        if cls.__name__ != 'Command':
            register_command(cls)
        return cls


class Command(metaclass=Meta):
    name = ''
    help = ''

    def setup(self, subparsers):
        parser = subparsers.add_parser(self.name, help=self.help)
        self.configure_parser(parser)

    def configure_parser(self, parser):
        pass

    def __call__(self, args):
        pass



from .tokenizer import CharacterTokenizer
from torch.nn.functional import softmax


def decode_output_batch(tensor, tokenizer):
    """Convert tensor of predictions into a list of strings.

    :param tensor: tensor of shape (batch_size, max_steps, num_classes) containing
    raw (unnormalized) scores representing how likely is a given character at a given step

    :param tokenizer: instance of CharacterTokenizer class
    :return: textual transcripts extracted using predictions tensor
    """

    transcripts, _ = decode_and_score(tensor, tokenizer)
    return transcripts


def decode_and_score(tensor, tokenizer):
    """Convert tensor of predictions into a list of strings and compute confidence scores.

    :param tensor: tensor of shape (batch_size, max_steps, num_classes) containing
    raw (unnormalized) scores representing how likely is a given character at a given step

    :param tokenizer: instance of CharacterTokenizer class
    :return: 2-tuples with transcripts and corresponding confidence score
    """

    pmf = softmax(tensor, dim=2)
    values, indices = pmf.max(dim=2)

    end_token = tokenizer.encode(tokenizer.end)

    all_transcripts = []
    all_scores = []

    for i in range(len(indices)):
        tokens = indices[i].tolist()

        try:
            first_n = tokens.index(end_token)
        except ValueError:
            first_n = len(tokens)

        confidence_score = values[i, :first_n].mean()

        transcript = tokenizer.decode_to_string(tokens, clean_output=True)

        all_transcripts.append(transcript)
        all_scores.append(confidence_score)

    return all_transcripts, all_scores


class DecodeBatchTransform:
    def __init__(self, charset):
        self.tokenizer = CharacterTokenizer(charset)

    def __call__(self, y_hat, y):
        return decode_output_batch(y_hat, self.tokenizer), y



import json
import os
import shutil

import torch
from .configuration import Configuration


def clean_metrics(metrics):
    res = {}
    for name, value in metrics.items():
        if hasattr(value, 'item'):
            value = value.item()
        res[name] = value

    return res


class CheckpointKeeper:
    def __init__(self, checkpoints_dir):
        self.checkpoints_dir = checkpoints_dir

    def get_latest_meta_data(self):
        highest = self._get_highest_checkpoint_number()
        checkpoint_folder = os.path.join(self.checkpoints_dir, str(highest))
        return Checkpoint(checkpoint_folder).meta_data

    def load_latest_checkpoint(self, trainable, device):
        highest = self._get_highest_checkpoint_number()
        highest_dir = os.path.join(self.checkpoints_dir, str(highest))
        return self.load_checkpoint(trainable, highest_dir, device)

    def load_checkpoint(self, trainable, checkpoint_dir, device):
        checkpoint = Checkpoint(checkpoint_dir)

        session_state = checkpoint.get_session_state(device)

        trainable.encoder.load_state_dict(session_state.encoder)
        trainable.encoder.to(device)

        trainable.decoder.load_state_dict(session_state.decoder)
        trainable.decoder.to(device)

        trainable.encoder_optimizer.load_state_dict(session_state.encoder_optimizer)
        trainable.decoder_optimizer.load_state_dict(session_state.decoder_optimizer)

    def make_new_checkpoint(self, trainable, device, epoch, metrics):
        try:
            highest = self._get_highest_checkpoint_number()
            checkpoint_name = str(highest + 1)
        except CheckpointsNotFound:
            checkpoint_name = '0'

        checkpoint_path = os.path.join(self.checkpoints_dir, checkpoint_name)
        os.makedirs(checkpoint_path)
        Checkpoint.create(checkpoint_path, trainable, device, epoch, metrics)

    def _get_highest_checkpoint_number(self):
        checkpoints = []
        for folder in os.listdir(self.checkpoints_dir):
            try:
                checkpoints.append(int(folder))
            except ValueError:
                pass

        if not checkpoints:
            raise CheckpointsNotFound()

        return max(checkpoints)

    def get_latest_checkpoint_dir(self):
        highest = self._get_highest_checkpoint_number()
        checkpoint_name = str(highest)
        return os.path.join(self.checkpoints_dir, checkpoint_name)


class Checkpoint:
    def __init__(self, folder):
        self._folder = folder
        meta_path = self._metadata_path(self._folder)

        with open(meta_path, encoding='utf-8') as f:
            self._meta_data = json.loads(f.read())

    @classmethod
    def create(cls, save_dir, trainable, device, epoch, metrics):
        checkpoint_path = cls._state_path(save_dir)
        meta_path = cls._metadata_path(save_dir)

        with open(meta_path, 'w', encoding='utf-8') as f:
            metrics = clean_metrics(metrics)
            meta_data = {
                'device': str(device),
                'epoch': epoch,
                'metrics': metrics
            }
            f.write(json.dumps(meta_data))

        models_dict = dict(encoder=trainable.encoder.state_dict(), decoder=trainable.decoder.state_dict())

        optimizers_dict = {
            'encoder_optimizer': trainable.encoder_optimizer.state_dict(),
            'decoder_optimizer': trainable.decoder_optimizer.state_dict()
        }

        torch.save({
            'models': models_dict,
            'optimizers': optimizers_dict
        }, checkpoint_path)
        return cls(save_dir)

    @classmethod
    def _state_path(cls, folder):
        return os.path.join(folder, 'checkpoint.pt')

    @classmethod
    def _metadata_path(cls, folder):
        return os.path.join(folder, 'metadata.txt')

    def get_session_state(self, device):
        state_path = self._state_path(self._folder)
        if self.meta_data["device"] == device:
            state_dict = torch.load(state_path)
        else:
            state_dict = torch.load(state_path, map_location=device)
        return SessionState(state_dict)

    @property
    def meta_data(self):
        return self._meta_data


class SessionState:
    def __init__(self, state_dict):
        self.state_dict = state_dict

    @property
    def encoder(self):
        return self.state_dict["models"]["encoder"]

    @property
    def decoder(self):
        return self.state_dict["models"]["decoder"]

    @property
    def encoder_optimizer(self):
        return self.state_dict["optimizers"]["encoder_optimizer"]

    @property
    def decoder_optimizer(self):
        return self.state_dict["optimizers"]["decoder_optimizer"]


class SessionDirectoryLayout:
    def __init__(self, session_dir):
        self.session = session_dir

        self.checkpoints = os.path.join(session_dir, "checkpoints")
        self.tuning_checkpoints = os.path.join(session_dir, "tuning_checkpoints")

        self.history = os.path.join(session_dir, "history.csv")
        self.tuning_history = os.path.join(session_dir, "tuning_history.csv")

        self.model_spec = os.path.join(session_dir, "model_spec.json")

    def make_session_dir(self):
        """Creates session directory if it does not exist"""
        if not os.path.exists(self.session):
            os.makedirs(self.session)

    def make_checkpoints_dir(self):
        """Creates checkpoints subdirectory if it does not exist"""
        if not os.path.exists(self.checkpoints):
            os.makedirs(self.checkpoints)

    def remove_history(self):
        self._remove_history_file(self.history)

    def remove_tuning_history(self):
        self._remove_history_file(self.tuning_history)

    def tuning_checkpoint_exists(self):
        path = os.path.join(self.tuning_checkpoints, '0')
        return os.path.exists(path)

    def create_tuning_checkpoint(self):
        """Copy checkpoint of a pretrained model into a directory for tuning checkpoints
        (only if this hasn't been already done)
        """
        if self.tuning_checkpoint_exists():
            return

        keeper = CheckpointKeeper(self.checkpoints)
        checkpoint_path = keeper.get_latest_checkpoint_dir()
        dest_path = os.path.join(self.tuning_checkpoints, '0')
        shutil.copytree(checkpoint_path, dest_path)

        self.update_meta_data(dest_path)

        self.remove_tuning_history()

    def update_meta_data(self, checkpoint_dir):
        tuning_keeper = CheckpointKeeper(self.tuning_checkpoints)
        meta_data = tuning_keeper.get_latest_meta_data()
        meta_data["epoch"] = 0

        # todo: change extension to json
        meta_path = os.path.join(checkpoint_dir, "metadata.txt")
        meta_json = json.dumps(meta_data)
        with open(meta_path, "w") as f:
            f.write(meta_json)

    def _remove_history_file(self, history_path):
        if os.path.isfile(history_path):
            os.remove(history_path)

    def load_config(self):
        config_path = os.path.join(self.session, "config.json")
        with open(config_path) as f:
            json_str = f.read()

        return load_conf(json_str)


def load_conf(json_str):
    d = json.loads(json_str)
    obj = Configuration()
    for name, value in d['fields'].items():
        setattr(obj, name, value)

    obj.device = torch.device(obj.device)
    return obj


class CheckpointsNotFound(Exception):
    """Raised when trying to load a checkpoint from a folder containing none of them"""



from dataclasses import dataclass
from typing import Any, List, AnyStr

import torch
from torch.utils.data import DataLoader

from lafhterlearn.formatters import Formatter
from lafhterlearn.metrics import MetricsSetCalculator, Metric
from lafhterlearn.recognition import WordRecognitionPipeline
from lafhterlearn.tokenizer import CharacterTokenizer
from lafhterlearn.augmentation import WeakAugmentation, StrongAugmentation


def print_metrics(computed_metrics, epoch):
    formatter = Formatter()

    final_metrics_string = formatter.format_metrics(computed_metrics)

    epoch_str = formatter.format_epoch(epoch)

    whitespaces = ' ' * 150
    print(f'\r{whitespaces}\r{epoch_str} {final_metrics_string}')


@dataclass
class IterationLogEntry:
    iteration: int
    num_iterations: int
    outputs: torch.Tensor
    targets: List[AnyStr]
    loss: Any


@dataclass
class BaseTrainer:
    recognizer: WordRecognitionPipeline
    data_loader: DataLoader
    loss_fn: Metric
    tokenizer: CharacterTokenizer
    supress_errors: bool = True

    def __iter__(self):
        num_iterations = len(self.data_loader)
        self.recognizer.neural_pipeline.train_mode()

        for i, (images, transcripts) in enumerate(self.data_loader):
            try:
                loss, result = self.train_one_iteration(images, transcripts)
            except torch.cuda.OutOfMemoryError:
                if not self.supress_errors:
                    raise
                continue

            outputs = result
            targets = transcripts
            yield IterationLogEntry(i, num_iterations, outputs, targets, loss)

    def train_one_iteration(self, images, transcripts):
        loss, y_hat = self.compute_loss(images, transcripts)

        # invoke zero_grad for each neural network
        self.recognizer.neural_pipeline.zero_grad()
        loss.backward()

        # todo: clip gradients

        # invoke optimizer.step() for every neural network if there is one
        self.recognizer.neural_pipeline.step()

        return loss, y_hat

    def compute_loss(self, images, transcripts):
        raise NotImplemented


class Trainer(BaseTrainer):
    def compute_loss(self, images, transcripts):
        y_hat = self.recognizer(images, transcripts)
        loss = self.loss_fn(y_hat=y_hat, y=transcripts)
        return loss, y_hat


class ConsistencyTrainer(BaseTrainer):
    def __init__(self, recognizer, data_loader, loss_fn, tokenizer, weak_augment, strong_augment):
        super().__init__(recognizer, data_loader, loss_fn, tokenizer)
        self.weak_augment = weak_augment
        self.strong_augment = strong_augment

    def compute_loss(self, images, transcripts):
        y_weak = self.recognizer(self.weak_augment(images), transcripts)
        y_strong = self.recognizer(self.strong_augment(images), transcripts)

        loss_weak = self.loss_fn(y_hat=y_weak, y=transcripts)
        loss_strong = self.loss_fn(y_hat=y_strong, y=transcripts)
        loss = loss_weak + loss_strong
        return loss, y_weak


@dataclass
class TrainingLoop:
    trainer: BaseTrainer
    metric_fns: dict
    epochs: int
    starting_epoch: int

    def __iter__(self):
        for epoch in range(self.starting_epoch, self.epochs + 1):
            calculator = MetricsSetCalculator(self.metric_fns, interval=100)

            for iteration_log in self.trainer:
                running_metrics = calculator(iteration_log)
                self.print_metrics(epoch, iteration_log, running_metrics)

            yield epoch

    def print_metrics(self, epoch, iteration_log: IterationLogEntry, running_metrics):
        formatter = Formatter()

        stats = formatter(
            epoch, iteration_log.iteration + 1,
            iteration_log.num_iterations, running_metrics
        )
        print(f'\r{stats}', end='')


def get_simple_trainer(recognizer, loader, loss_fn, tokenizer, **kwargs):
    return Trainer(recognizer, loader, loss_fn, tokenizer)


def get_consistency_trainer(recognizer, loader, loss_fn, tokenizer,
                            **weak_augment_options):
    weak_augment = WeakAugmentation(**weak_augment_options)
    strong_augment = StrongAugmentation()
    return ConsistencyTrainer(recognizer, loader, loss_fn, tokenizer,
                              weak_augment, strong_augment)



import logging.config

logger = logging.getLogger(__name__)

logger.setLevel(logging.DEBUG)

handler = logging.FileHandler('lafhterlearn.log')
handler.setLevel(logging.DEBUG)

# create formatter
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

# add formatter to ch
handler.setFormatter(formatter)

# add ch to logger
logger.addHandler(handler)



import os
import csv


class HistoryCsvSaver:
    def __init__(self, file_path):
        self.file_path = file_path

    def add_entry(self, epoch, metrics):
        all_metrics = {}
        all_metrics.update(metrics)

        row_dict = {'epoch': epoch}
        row_dict.update({k: self.scalar(v) for k, v in all_metrics.items()})

        field_names = list(row_dict.keys())

        if not os.path.exists(self.file_path):
            self.create(self.file_path, field_names)

        with open(self.file_path, 'a', encoding='utf-8', newline='') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=field_names)
            writer.writerow(row_dict)

    def scalar(self, t):
        return t.item() if hasattr(t, 'item') else t

    @classmethod
    def create(cls, path, field_names):
        with open(path, 'w', encoding='utf-8', newline='') as csvfile:
            writer = csv.writer(csvfile)
            writer.writerow(field_names)
        return cls(path)



import os
import random
import logging

from PIL import Image, ImageDraw, ImageFont
from torchvision import transforms


logger = logging.getLogger(__name__)


class SimpleRandomWordGenerator:
    def __init__(self, word_sampler, font_dir, font_size_range=(64, 64),
                 bg_range=(240, 255), color_range=(0, 15), stroke_width_range=(0, 2),
                 stroke_fill_range=(0, 15), rotation_range=(0, 0), spaces_range=(0, 2)):
        self.sampler = word_sampler
        self.font_dir = font_dir
        self.font_files = [os.path.join(font_dir, font_file) for font_file in os.listdir(font_dir)
                           if font_file.endswith('.otf') or font_file.endswith('.ttf')]

        self.font_size_range = font_size_range
        self.bg_range = bg_range
        self.color_range = color_range
        self.stroke_width_range = stroke_width_range
        self.stroke_fill_range = stroke_fill_range
        self.rotation_range = rotation_range
        self.spaces_range = spaces_range

    def __iter__(self):
        while True:
            word = self.sampler()

            font_size = random.randint(*self.font_size_range)
            font_file = random.choice(self.font_files)
            font = ImageFont.truetype(font_file, size=font_size)

            background = random.randint(*self.bg_range)
            color = random.randint(*self.color_range)
            stroke_fill = random.randint(*self.stroke_fill_range)
            stroke_width = random.randint(*self.stroke_width_range)
            num_spaces = random.randint(*self.spaces_range)

            word_with_spaces = add_spacing(word, num_spaces)
            try:
                image = self.create_image(word_with_spaces, font, font_size,
                                          background=background, color=color,
                                          stroke_width=stroke_width, stroke_fill=stroke_fill)
                if image.height > 0 and image.width > 0:
                    yield image, word
            except Exception:
                msg = 'Failed to create image for "{}": font "{}", font size {}, ' \
                      'background {}, color {}, stroke width {}, stroke fill {}, # spaces {}'

                # todo: this is wrong when using extra worker processes in dataloader
                logger.exception(msg.format(
                    word, font_file, font_size, background, color,
                    stroke_width, stroke_fill, num_spaces
                ))

    def create_image(self, word, font, size=64, background=255, color=0,
                     stroke_width=1, stroke_fill=0):
        padding = stroke_width
        char_size = size
        num_chars = len(word)
        width = char_size * num_chars + padding * 2

        vertical_offset = 25
        height = size * 2 + 20 + padding * 2 + vertical_offset * 2

        min_degrees, max_degrees = self.rotation_range
        rotate = transforms.RandomRotation(degrees=[min_degrees, max_degrees], expand=True, fill=background)

        with Image.new("L", (width, height)) as image:
            draw = ImageDraw.Draw(image)

            # give extra vertical space (at least 25 pixels from top and bottom)
            bbox = draw.textbbox((padding, vertical_offset), word, font=font)
            draw.rectangle((0, 0, image.width, image.height), fill=background)
            draw.text((padding, vertical_offset), word, fill=color, font=font,
                      stroke_width=stroke_width, stroke_fill=stroke_fill)

            x0, y0, x, y = bbox
            padded_bbox = (max(0, x0 - padding),
                           max(0, y0 - padding),
                           min(width, x + padding),
                           min(height, y + padding))

            shear_x = transforms.RandomAffine(0, shear=(-10, 30), fill=background)

            image = image.crop(padded_bbox)
            image = shear_x(image)

            if self.rotation_range != (0, 0):
                image = rotate(image)
            return image


def add_spacing(s, n=0, space=0x202F):
    narrow_space = chr(space) * n
    return narrow_space.join(s)



import torch

from collections import namedtuple
from .metrics import MovingAverage, update_running_metrics
from .formatters import show_progress_bar


def evaluate(task, supress_errors=True):
    """Compute metrics.

    :param task: instance of EvaluationTask
    :param supress_errors: When set to True (by default), it will silently catch exceptions
    such as torch.cuda.OutOfMemoryError. Otherwise, exceptions will propagate to the caller.
    :return: a dictionary mapping names of metrics to their computed values
    """
    recognizer = task.recognizer
    recognizer.neural_pipeline.eval_mode()

    metrics = task.metric_functions
    data_generator = task.data_loader

    num_batches = calculate_num_batches(task.num_batches, len(data_generator))

    moving_averages = {name: MovingAverage() for name in metrics}

    gen = show_progress_bar(data_generator, desc='Evaluating metrics: ',
                            num_iters=num_batches, cols=50)
    with torch.no_grad():
        for i, (images, transcripts) in enumerate(gen):
            if i >= num_batches:
                break

            args = [images] if task.close_loop_prediction else [images, transcripts]

            try:
                y_hat = recognizer(*args)
                d = dict(y=transcripts, y_hat=y_hat)
                update_running_metrics(moving_averages, metrics, d)
            except torch.cuda.OutOfMemoryError:
                if not supress_errors:
                    raise

    return {metric_name: avg.value for metric_name, avg in moving_averages.items()}


def calculate_num_batches(num_batches, batches_total):
    if isinstance(num_batches, float):
        fraction = num_batches
        num_batches = int(round(batches_total * fraction))
        num_batches = max(1, num_batches)
    return num_batches


EvaluationTask = namedtuple('EvaluationTask',
                            ['recognizer', 'data_loader', 'metric_functions',
                             'num_batches', 'close_loop_prediction'])



import torch


class MetricsSetCalculator:
    def __init__(self, metrics, interval):
        """
        :param metrics: {'name': ('graph_leaf', Metric())}
        """
        self.metrics = metrics
        self.interval = interval
        self.running_metrics = {name: MovingAverage() for name in metrics}

    def __call__(self, iteration_log):
        if iteration_log.iteration % self.interval == 0:
            self.reset()

        with torch.no_grad():
            d = dict(y=iteration_log.targets, y_hat=iteration_log.outputs)
            update_running_metrics(self.running_metrics, self.metrics, d)

        return self.running_metrics

    def reset(self):
        for metric_avg in self.running_metrics.values():
            metric_avg.reset()


class Metric:
    def __init__(self, name, metric_fn, metric_args, transform_fn=None):
        """

        :param name: A string that identifies the metric
        :param metric_fn: a callable that maps variable number of tensors to a scalar
        :param metric_args: tensor names of batch that are used to compute the metric on
        :type metric_args: iterable of type str
        :param transform_fn: a callable that transforms/reduces # of the metrics arguments
        """
        self.name = name
        self.metric_fn = metric_fn
        self.metric_args = metric_args
        self.transform_fn = transform_fn or identity

    def rename_and_clone(self, new_name):
        return self.__class__(new_name, self.metric_fn, self.metric_args, self.transform_fn)

    def __call__(self, **batch):
        """Given the batch of predictions and targets, compute the metric.

        The method will first extract all arguments that metric
        function expects from batch. Then it will apply transforms to them.
        Finally, it will place all tensors on a specified device and compute the metric value.

        Note that this method will also compute gradients with respect to tensors that participate in
        the metrics calculation. If this is not required, one should manually detach() tensors or
        make the call under no_grad() context manager.

        :param batch: a dictionary mapping a name of a metric argument to a corresponding tensor
        :type batch: Dict[str -> torch.tensor]
        :return: a metric scalar
        :rtype: degenerate tensor of shape ()
        """

        lookup_table = batch
        inputs = [lookup_table[arg] for arg in self.metric_args]

        device = self.fastest_device(inputs)

        inputs = self.change_device(inputs, device)
        inputs = self.transform_fn(*inputs)
        inputs = self.change_device(inputs, device)  # the above operation could change devices

        fn = self.metric_fn.to(device) if hasattr(self.metric_fn, 'to') else self.metric_fn
        return fn(*inputs)

    def fastest_device(self, inputs):
        """Returns cuda device if at least 1 input tensor is on cuda,
        otherwise return cpu device"""
        devices = [v.device for v in inputs if hasattr(v, 'device')]

        for d in devices:
            if 'cuda' in str(d):
                return d
        return torch.device('cpu')

    def change_device(self, tensors, device):
        """Moves all tensors that participate in metric calculation to a given device

        :param tensors: a list of tensors
        :param device: torch.device instance
        :return: a list of tensors
        """

        # some tensor may already be on the right device, if so they kept unchanged
        return [arg if not hasattr(arg, 'device') or arg.device == device else arg.to(device)
                for arg in tensors]


def identity(*inputs):
    return inputs


class MovingAverage:
    def __init__(self):
        self.x = 0
        self.num_updates = 0

    def reset(self):
        self.x = 0
        self.num_updates = 0

    def update(self, x):
        self.x += x
        self.num_updates += 1

    @property
    def value(self):
        return self.x / self.num_updates


def update_running_metrics(moving_averages, metrics, results_batch):
    for name, metric_fn in metrics.items():
        moving_averages[name].update(
            metric_fn(**results_batch)
        )



import os

from PIL import ImageFont

from lafhterlearn.word_samplers import UniformSampler
from lafhterlearn.data_generator import SimpleRandomWordGenerator, add_spacing
from .base import Command


class VisualizeFontsCommand(Command):
    name = 'visualize_fonts'
    help = 'Create text images using every font in a given fonts directory'

    def configure_parser(self, parser):
        parser.add_argument('fonts_dir', type=str,
                            help='Path to the directory containing files with .ttf or .otf extension ')
        parser.add_argument('output_dir', type=str,
                            help='Output directory that will contain generated images')

        parser.add_argument('--text', type=str, default='The quick brown fox jumps over the lazy dog',
                            help='Specifies a text to write using each font')

        parser.add_argument('--font-size', type=int, default=64,
                            help='Font size to use with each font')

        parser.add_argument('--max-fonts', type=int, default=0,
                            help='Number of fonts to visualize. By default, visualize all')

        parser.add_argument('--num-spaces', type=int, default=0,
                            help='Number of narrow spaces between letters')

    def __call__(self, args):
        generate_images(args)


def generate_images(args):
    fonts_dir = args.fonts_dir
    output_dir = args.output_dir
    text = args.text
    font_size = args.font_size
    max_fonts = args.max_fonts
    num_spaces = args.num_spaces

    if num_spaces > 0:
        text = add_spacing(text, num_spaces)

    sampler = UniformSampler([])
    gen = SimpleRandomWordGenerator(sampler, fonts_dir)

    os.makedirs(output_dir, exist_ok=True)

    for i, font_name in enumerate(os.listdir(fonts_dir)):
        if max_fonts and i >= max_fonts:
            break

        if font_name.endswith('.otf') or font_name.endswith('.ttf'):
            font_path = os.path.join(fonts_dir, font_name)
            font = ImageFont.truetype(font_path, size=font_size)
            save_path = os.path.join(output_dir, f"{i}_{font_name}.png")
            gen.create_image(text, font, size=font_size, stroke_width=0).save(save_path)
        else:
            print(f"Ignoring file with wrong extension: {font_name}")



import json
import os
import importlib

from torch.utils.data import DataLoader

from lafhterlearn.utils import collate, split_import_path

from .models import build_from_spec

from .recognition import (
    WordRecognitionPipeline,
    TrainableEncoderDecoder
)
from .image_pipelines import make_pretraining_pipeline, make_validation_pipeline
from .datasets import (
    SyntheticOnlineDataset,
    SyntheticOnlineDatasetCached,
    RealUnlabeledDataset,
    RealLabeledDataset
)

from .session import (
    CheckpointKeeper,
    SessionDirectoryLayout,
    CheckpointsNotFound
)
from .training import (
    TrainingLoop,
    Trainer,
    get_simple_trainer,
    get_consistency_trainer
)
from .history_saver import HistoryCsvSaver
from .evaluation import EvaluationTask

from .config_utils import (
    prepare_metrics,
    prepare_loss,
    create_optimizer
)

from .configuration import Configuration

from .tokenizer import CharacterTokenizer


def create_neural_pipeline(device, model_spec, config):
    encoder, decoder = build_from_spec(model_spec)

    encoder_optimizer = create_optimizer(encoder, config.encoder_optimizer)
    decoder_optimizer = create_optimizer(decoder, config.decoder_optimizer)
    return TrainableEncoderDecoder(
        encoder, decoder, device, encoder_optimizer, decoder_optimizer
    )


def load_neural_pipeline(config, session_layout):
    """Instantiate encoder-decoder model and restore it from checkpoint if it exists,
    otherwise create checkpoint.

    Returns encoder-decoder model
    """

    spec = load_model_spec(session_layout)

    neural_pipeline = create_neural_pipeline(config.device, spec, config)

    keeper = CheckpointKeeper(session_layout.checkpoints)

    keeper.load_latest_checkpoint(neural_pipeline, config.device)
    return neural_pipeline


def load_model_spec(session_layout):
    with open(session_layout.model_spec) as f:
        s = f.read()
        return json.loads(s)


class Environment:
    def __init__(self, config: Configuration):
        self.config = config
        session_layout = SessionDirectoryLayout(config.session_dir)

        self.session_layout = session_layout

        tokenizer = CharacterTokenizer(config.charset)

        neural_pipeline = load_neural_pipeline(config, session_layout)

        image_pipeline = make_pretraining_pipeline(
            augmentation_options=config.weak_augment_options,
            max_width=config.max_image_width,
            max_height=config.image_height,
            resize=config.resize_images
        )
        recognizer = WordRecognitionPipeline(neural_pipeline, tokenizer, image_pipeline)

        loss_fn = prepare_loss(config.loss_function)

        train_metric_fns = prepare_metrics(config.training_metrics)
        val_metric_fns = prepare_metrics(config.validation_metrics)
        test_metric_fns = prepare_metrics(config.test_metrics)

        training_loader = self._create_data_loader(SyntheticOnlineDataset,
                                                   config.training_set_size)

        val_loader = self._create_data_loader(SyntheticOnlineDatasetCached,
                                              config.validation_set_size)

        trainer = Trainer(recognizer, training_loader, loss_fn, tokenizer)

        epochs_trained = self._get_trained_epochs()

        training_loop = TrainingLoop(trainer, metric_fns=train_metric_fns,
                                     epochs=config.epochs,
                                     starting_epoch=epochs_trained + 1)

        eval_on_train = EvaluationTask(recognizer, training_loader, train_metric_fns,
                                       config.evaluation_steps['training_set'],
                                       close_loop_prediction=False)

        val_preprocessor = make_validation_pipeline(max_width=config.max_image_width,
                                                    max_height=config.image_height,
                                                    resize=config.resize_images)

        val_recognizer = WordRecognitionPipeline(neural_pipeline, tokenizer, val_preprocessor)
        eval_on_val = EvaluationTask(val_recognizer, val_loader, val_metric_fns,
                                     config.evaluation_steps['validation_set'],
                                     close_loop_prediction=False)

        self.neural_pipeline = neural_pipeline
        self.epochs_trained = epochs_trained
        self.training_loop = training_loop
        self.history_saver = HistoryCsvSaver(session_layout.history)
        self.eval_tasks = [eval_on_train, eval_on_val]

        test_ds_path = os.path.join(config.tuning_data_dir, "labeled")

        if os.path.isdir(test_ds_path):
            test_ds = RealLabeledDataset(test_ds_path)
            test_loader = DataLoader(test_ds, batch_size=config.batch_size,
                                     num_workers=config.num_workers, collate_fn=collate)

            for name, metric_fn in test_metric_fns.items():
                close_loop_on = config.close_loop_mode.get(name, False)
                task = EvaluationTask(val_recognizer, test_loader, {name: metric_fn},
                                      num_batches=config.evaluation_steps['test_set'],
                                      close_loop_prediction=close_loop_on)
                self.eval_tasks.append(task)

    def save_checkpoint(self, epoch, metrics):
        keeper = CheckpointKeeper(self.session_layout.checkpoints)
        keeper.make_new_checkpoint(self.neural_pipeline, self.config.device, epoch, metrics)

    def _get_trained_epochs(self):
        try:
            keeper = CheckpointKeeper(self.session_layout.checkpoints)
            meta_data = keeper.get_latest_meta_data()
            return meta_data["epoch"]
        except CheckpointsNotFound:
            return 0

    def _create_data_loader(self, dataset_class, dataset_size):
        module_path, class_name = split_import_path(self.config.word_sampler)
        module = importlib.import_module(module_path)
        cls = getattr(module, class_name)

        sampler = cls.from_file(self.config.sampler_data_file)
        ds = dataset_class(
            self.config.fonts_dir, dataset_size,
            word_sampler=sampler,
            **self.config.data_generator_options
        )

        return DataLoader(ds, batch_size=self.config.batch_size,
                          num_workers=self.config.num_workers, collate_fn=collate)


class TuningEnvironment:
    def __init__(self, config: Configuration):
        self.config = config

        unlabeled_ds_path = os.path.join(config.tuning_data_dir, "unlabeled")
        unlabeled_ds = RealUnlabeledDataset(unlabeled_ds_path)
        unlabeled_loader = self._create_loader(unlabeled_ds)

        session_layout = SessionDirectoryLayout(config.session_dir)
        session_layout.create_tuning_checkpoint()
        self.session_layout = session_layout

        tokenizer = CharacterTokenizer(config.charset)

        model_spec = load_model_spec(session_layout)

        encoder_decoder = create_neural_pipeline(config.device, model_spec, config)
        keeper = CheckpointKeeper(session_layout.tuning_checkpoints)
        keeper.load_latest_checkpoint(encoder_decoder, config.device)

        image_preprocessor = make_validation_pipeline(max_width=config.max_image_width,
                                                      max_height=config.image_height,
                                                      resize=config.resize_images)
        recognizer = WordRecognitionPipeline(encoder_decoder, tokenizer, image_preprocessor)

        metric_fns = prepare_metrics(config.training_metrics)

        test_metrics = prepare_metrics(config.test_metrics)

        eval_steps = config.evaluation_steps

        self.unlabeled_loader = unlabeled_loader
        self.tokenizer = tokenizer
        self.threshold = config.confidence_threshold
        self.tuning_epochs = config.tuning_epochs
        self.neural_pipeline = encoder_decoder
        self.recognizer = recognizer
        self.metric_fns = metric_fns

        self.tasks = []

        self.history_saver = HistoryCsvSaver(session_layout.tuning_history)

        test_ds_path = os.path.join(config.tuning_data_dir, "labeled")

        if os.path.isdir(test_ds_path):
            test_ds = RealLabeledDataset(test_ds_path)
            test_loader = self._create_loader(test_ds)

            for name, metric_fn in test_metrics.items():
                close_loop_on = config.close_loop_mode.get(name, False)
                task = EvaluationTask(recognizer, test_loader, {name: metric_fn},
                                      num_batches=eval_steps["test_set"],
                                      close_loop_prediction=close_loop_on)
                self.tasks.append(task)

    def _create_loader(self, ds):
        return DataLoader(ds,
                          batch_size=self.config.batch_size,
                          num_workers=self.config.num_workers,
                          collate_fn=collate)

    def get_trainer(self, pseudo_labeled_ds):
        loader = self._create_loader(pseudo_labeled_ds)

        loss_fn = prepare_loss(self.config.loss_function)

        tokenizer = CharacterTokenizer(self.config.charset)

        if self.config.tuning_trainer_factory == "simple_trainer":
            trainer_factory = get_simple_trainer
        else:
            trainer_factory = get_consistency_trainer

        return trainer_factory(
            self.recognizer, loader, loss_fn, tokenizer,
            **self.config.weak_augment_options)

    def save_checkpoint(self, epoch, metrics):
        keeper = CheckpointKeeper(self.session_layout.tuning_checkpoints)
        keeper.make_new_checkpoint(self.neural_pipeline, self.config.device, epoch, metrics)

    def get_trained_epochs(self):
        keeper = CheckpointKeeper(self.session_layout.tuning_checkpoints)
        meta_data = keeper.get_latest_meta_data()
        return meta_data["epoch"]



import os

import nltk
from .base import Command
from lafhterlearn.ngrams import build_ngram_model


class CreateWordDistributionCommand(Command):
    name = 'learn_ngrams'
    help = 'Build n-grams using a plain text corpora'

    def configure_parser(self, parser):
        parser.add_argument('output_file', type=str, help='Location of output dictionary file')

        parser.add_argument('corpora_dir', type=str, default='',
                            help='Location of a directory containing plain text files')

        parser.add_argument('--n', type=int, default=2,
                            help='Order of n-gram model (1 for unigrams, 2 for bigrams, etc.). 2 by default')

        parser.add_argument('--max-words', type=int, default=10**12,
                            help='Maximum number of words to consider in text corpora before stopping. '
                                 'Consider all words by default')

        parser.add_argument('--only-letters', default=False, action='store_true',
                            help='Filters out words containing digits and other non-letter characters')

    def __call__(self, args):
        corpora_dir = args.corpora_dir
        max_words = args.max_words
        n = args.n

        def get_corpus():
            return sentences_stream(corpora_dir, max_words)
        return build_ngram_model(get_corpus, args.output_file, n=n)


def load_sentences(corpora_dir):
    """A generator yielding words contained in text files located within corpora_dir folder"""
    for text_file in os.listdir(corpora_dir):
        path = os.path.join(corpora_dir, text_file)
        with open(path) as f:
            text = f.read().strip()
            for sent in nltk.sent_tokenize(text):
                yield nltk.word_tokenize(sent)


def sentences_stream(corpora_dir, max_words):
    gen = load_sentences(corpora_dir)
    gen = data_limiter(gen, max_words)
    return with_progress_bar(gen)


def data_limiter(sentences, max_words):
    words = 0
    for sent in sentences:
        if words > max_words:
            break

        words += len(sent)
        yield sent


def with_progress_bar(it):
    for i, elem in enumerate(it):
        if i % 100 == 0:
            print(f'\rProcessed sentences: {i + 1}', end='')
        yield elem
    print()



import torch
from torch import nn
from torchvision.models import vgg19_bn
from .tokenizer import CharacterTokenizer


class VGG19Truncated(nn.Module):
    def __init__(self):
        super().__init__()
        original_vgg = vgg19_bn()
        self.truncated_vgg = nn.Sequential(*list(original_vgg.children())[:-2])

    def forward(self, x):
        return self.truncated_vgg(x)


class ImageEncoder(nn.Module):
    def __init__(self, image_height, hidden_size):
        super().__init__()

        out_feature_maps = 512  # vgg outputs 512 feature maps
        feature_height = image_height // 2**5
        input_size = feature_height * out_feature_maps
        self.hidden_size = hidden_size
        self.vgg = VGG19Truncated()
        self.gru1 = nn.GRU(input_size, hidden_size, bidirectional=True, batch_first=True)
        self.dropout = nn.Dropout(0.5)
        self.gru2 = nn.GRU(hidden_size * 2, hidden_size, bidirectional=True, batch_first=True)

    def forward(self, x):
        output = self.vgg(x)
        batch_size, f_maps, h, w = output.size()
        x = output.transpose(1, 3).resize(batch_size, w, h * f_maps)

        x, hidden = self.gru1(x)

        x = self.dropout(x)

        x, hidden = self.gru2(x)

        return x

    def run_inference(self, x):
        return self.forward(x)


class HybridAttention(nn.Module):
    # adapted from paper https://arxiv.org/pdf/1506.07503.pdf, equation (9)

    def __init__(self, embedding_size, decoder_state_dim, hidden_size, filters, kernel_size):
        """

        :param embedding_size: size of encoder embedding vectors
        :param decoder_state_dim: size of decoder hidden state
        :param hidden_size: number of units in a hidden layer (of a scoring network)
        :param filters: number of 1d convolutional filters
        :param kernel_size: kernel size of each convolutional filter
        """
        super().__init__()

        # corresponds to a matrix F used to convolve attention weights from a previous step
        self.conv1d = nn.Conv1d(1, filters, kernel_size=kernel_size, padding='same')

        # pack W, V, U, b into a single linear layer operating on vectors of size
        # embedding_size + hidden_size + filters
        self.linear1 = nn.Linear(
            embedding_size + decoder_state_dim + filters, hidden_size, bias=True
        )

        # w transpose from equation (9)
        self.linear2 = nn.Linear(hidden_size, 1, bias=False)

    def forward(self, decoder_hidden, encoder_outputs, attention_weights):
        """Computes attention context vectors as a weighted sum of encoder_outputs
        :param decoder_hidden: hidden state of the decoder
        :type decoder_hidden: tensor of shape (1, batch_size, num_cells)
        :param encoder_outputs: outputs of the encoder
        :type encoder_outputs: tensor of shape (batch_size, num_steps, embedding_size)
        :param attention_weights: attention weights computed on the previous time step
        :type attention_weights: tensor of shape (batch_size, num_embedding_vectors)
        :return: new attention weights
        :rtype: tensor of shape (batch_size, num_embedding_vectors)
        """

        embeddings = encoder_outputs
        batch_size, num_steps, embedding_size = embeddings.shape
        
        s = self.repeat_state(decoder_hidden, num_steps)
        
        features = self.convolve(self.conv1d, attention_weights)
        f = features.transpose(1, 2)

        x = self.concat_features(embeddings, s, f)

        x = torch.tanh(self.linear1(x))
        x = self.linear2(x)
        scores = x.squeeze(2)

        new_attention_weights = self.smoothing(scores)

        return new_attention_weights

    @staticmethod
    def convolve(conv1d, attention_weights):
        attention_weights = attention_weights.unsqueeze(1)
        return conv1d(attention_weights)

    @staticmethod
    def repeat_state(t, num_repeats):
        """Maps tensor of shape (1, B, n) to shape (B, num_repeats, n) where each step is repeated"""
        t = t.squeeze(0)
        t = t.unsqueeze(1)

        return t.repeat(1, num_repeats, 1)

    @staticmethod
    def concat_features(*args):
        return torch.cat(args, dim=2)

    @staticmethod
    def smoothing(scores):
        sigmoid_scores = torch.sigmoid(scores)
        _, n = sigmoid_scores.shape
        feature_sum = sigmoid_scores.sum(dim=1)
        feature_sum = feature_sum.unsqueeze(1).repeat(1, n)
        return sigmoid_scores / feature_sum

    @staticmethod
    def compute_context_vectors(new_attention_weights, embeddings):
        return torch.bmm(new_attention_weights.unsqueeze(1), embeddings).squeeze(1)

    @staticmethod
    def initial_attention_weights(batch_size, num_embeddings, device):
        attention_weights = list(
            reversed([i / num_embeddings for i in range(num_embeddings)])
        )
        attention_weights = torch.tensor(attention_weights, device=device)
        attention_weights = attention_weights.unsqueeze(0).repeat(batch_size, 1)

        return HybridAttention.smoothing(attention_weights)


class AttendingDecoder(nn.Module):
    def __init__(self, sos_token, context_size, y_size,
                 hidden_size=256, filters=10, kernel_size=5):
        super().__init__()

        # start of sequence token
        self.sos_token = sos_token

        self.y_size = y_size

        self.hidden_size = hidden_size

        self.attention = HybridAttention(
            embedding_size=context_size, decoder_state_dim=hidden_size,
            hidden_size=hidden_size, filters=filters, kernel_size=kernel_size
        )
        self.decoder_gru = nn.GRU(context_size + y_size, hidden_size, batch_first=True)

        self.linear = nn.Linear(hidden_size, y_size)

    def forward(self, encodings, y_shifted=None):
        decoder_hidden = torch.zeros(1, len(encodings), self.hidden_size, device=encodings.device)

        if y_shifted is None:
            # close loop inference
            sos = torch.zeros(len(encodings), self.y_size, dtype=encodings.dtype, device=encodings.device)
            sos[:, self.sos_token] = 1
            return self.close_loop_inference(encodings, decoder_hidden, sos)

        # teacher forcing

        batch_size, num_steps, num_classes = y_shifted.size()
        num_embeddings = encodings.shape[1]

        outputs = []

        # initializing attention weights (higher values on the left end, lower ones on the right end)
        attention_weights = self.attention.initial_attention_weights(
            batch_size, num_embeddings, device=encodings.device
        )

        for t in range(num_steps):
            y_hat_prev = y_shifted[:, t, :]
            log_pmf, decoder_hidden, attention_weights = self.predict_next(
                decoder_hidden, encodings, attention_weights, y_hat_prev
            )
            outputs.append(log_pmf)

        y_hat = torch.stack(outputs, dim=1)
        return y_hat

    def close_loop_inference(self, encodings, decoder_hidden, sos):
        outputs, _ = self._do_inference(encodings, decoder_hidden, sos)
        return outputs

    def debug_attention(self, encodings):
        decoder_hidden = torch.zeros(1, len(encodings), self.hidden_size, device=encodings.device)
        sos = torch.zeros(len(encodings), self.y_size, dtype=encodings.dtype, device=encodings.device)
        sos[:, self.sos_token] = 1
        return self._do_inference(encodings, decoder_hidden, sos)

    def _do_inference(self, encodings, decoder_hidden, sos, max_steps=15):
        """Generates sequence of tokens in autoregressive way
        Returns both predicted sequence of tokens and list of attention weight tensors,
        one tensor for each predicted token
        """

        # todo: consider to remove sos from the output (teacher forcing forward seems not to include it)
        outputs = [sos]

        y_hat_prev = sos

        batch_size = len(encodings)
        batch_indices = range(batch_size)

        _, num_embeddings, _ = encodings.shape
        attention_weights = self.attention.initial_attention_weights(
            batch_size, num_embeddings, device=encodings.device
        )

        attention_per_step = [attention_weights]
        for t in range(max_steps):
            scores, decoder_hidden, attention_weights = self.predict_next(
                decoder_hidden, encodings, attention_weights, y_hat_prev
            )

            top = scores.argmax(dim=1)

            y_hat_prev = torch.zeros(batch_size, self.y_size, device=encodings.device)
            y_hat_prev[batch_indices, top] = 1.0
            outputs.append(scores)

            attention_per_step.append(attention_weights)

        return torch.stack(outputs, dim=1), attention_per_step

    def predict_next(self, decoder_hidden, encoder_outputs, prev_attention_weights, y_hat_prev):
        new_attention_weights = self.attention(decoder_hidden, encoder_outputs, prev_attention_weights)
        c = self.attention.compute_context_vectors(new_attention_weights, encoder_outputs)

        v = torch.cat([y_hat_prev, c], dim=1).unsqueeze(1)

        h, hidden = self.decoder_gru(v, decoder_hidden)
        h = h.squeeze(1)
        return self.linear(h), hidden, new_attention_weights


def build_networks_spec(charset, image_height=64, hidden_size=128,
                        decoder_hidden_size=128, filters=10, kernel_size=5):
    context_size = hidden_size * 2

    tokenizer = CharacterTokenizer(charset)
    sos_token = tokenizer.char2index[tokenizer.start]

    return {
        'encoder': dict(image_height=image_height, hidden_size=hidden_size),
        'decoder': dict(sos_token=sos_token, context_size=context_size,
                        y_size=tokenizer.charset_size, hidden_size=decoder_hidden_size,
                        filters=filters, kernel_size=kernel_size)
    }


def build_from_spec(spec):
    encoder_spec = spec['encoder']
    decoder_spec = spec['decoder']

    encoder = ImageEncoder(**encoder_spec)
    decoder = AttendingDecoder(**decoder_spec)
    return encoder, decoder



import string
import json
import torch
from lafhterlearn.utils import full_class_name


optimizer_conf = {
    'class': 'torch.optim.Adam',
    'kwargs': dict(lr=0.0001)
}

charset = string.ascii_letters

loss_conf = {
    'class': 'lafhterlearn.loss_functions.MaskedCrossEntropy',
    'kwargs': dict(reduction='sum', label_smoothing=0.6),
    'transform': {
        'class': 'lafhterlearn.loss_functions.LossTargetTransform',
        'kwargs': dict(charset=charset)
    }
}

cer_conf = {
    'class': 'torchmetrics.CharErrorRate',
    'transform': {
        'class': 'lafhterlearn.decoding.DecodeBatchTransform',
        'kwargs': dict(charset=charset)
    }
}

device = 'cuda:0' if torch.cuda.is_available() else 'cpu'


class Configuration:
    def __init__(self):
        #self.max_image_width = 100 * 15
        self.max_image_width = None

        self.resize_images = False
        self.image_height = 64
        self.hidden_size = 128

        self.decoder_params = dict(
            decoder_hidden_size=self.hidden_size,
            filters=10,
            kernel_size=5
        )

        self.charset = charset

        self.fonts_dir = './fonts'

        self.data_generator_options = dict(
            bg_range=(200, 255),
            color_range=(0, 50),
            stroke_fill_range=(0, 50),
            font_size_range=(50, 100),
            rotation_range=(0, 0),
            spaces_range=(0, 2)
        )

        self.training_set_size = 50000
        self.validation_set_size = 2500

        self.batch_size = 32
        self.num_workers = 0

        self.loss_function = loss_conf

        self.encoder_optimizer = optimizer_conf
        self.decoder_optimizer = optimizer_conf

        self.training_metrics = {
            'loss': loss_conf,
            'CER': cer_conf
        }

        # evaluated without using augmentation
        self.validation_metrics = {
            'val loss': loss_conf,
            'val CER': cer_conf
        }

        # evaluated on test dataset with distribution close to one seen in production (without augmentation)
        self.test_metrics = {
            'test loss': loss_conf,
            'test CER': cer_conf
        }

        # whether to run inference in close loop mode when calculating test metrics
        self.close_loop_mode = {
            'test loss': False,
            'test CER': True
        }

        self.device = 'cuda:0' if torch.cuda.is_available() else 'cpu'

        self.session_dir = 'session'

        self.evaluation_steps = {
            'training_set': 0.1,
            'validation_set': 1.0,
            'test_set': 0.5
        }
        self.epochs = 50
        self.tuning_epochs = 50

        self.word_sampler = 'lafhterlearn.word_samplers.FrequencyBasedSampler'
        self.sampler_data_file = "word_distribution.csv"

        self.tuning_data_dir = 'tuning_data'
        self.confidence_threshold = 0.4

        self.tuning_trainer_factory = "simple_trainer"

        self.weak_augment_options = dict(
            p_augment=0.4,
            target_height=64,
            fill=255,
            rotation_degrees_range=(-5, 5),
            blur_size=3,
            blur_sigma=(1, 1),
            noise_sigma=10,
            should_fit_height=False
        )

    def to_json(self):
        d = {
            'class': full_class_name(self),
            'fields': self.__dict__
        }
        return json.dumps(d)



import os
import shutil

from .base import Command


class FontsCommand(Command):
    name = 'extract_fonts'
    help = 'Extracts pseudo-handwritten fonts from Google fonts folder'

    def configure_parser(self, parser):
        configure_parser(parser)

    def __call__(self, args):
        run(args)


def configure_parser(parser):
    parser.add_argument('fonts_dir', type=str,
                        help='Path to the location of google fonts directory')
    parser.add_argument('output_dir', type=str,
                        help='Path to the output directory')
    parser.add_argument('--num_fonts', type=int,
                        default=100, help='Maximum number of fonts to take')
    parser.add_argument(
        '--lang', type=str, default='latin',
        help='Specifies target language that fonts are required to support'
    )


def run(args):
    ofl_dir = os.path.join(args.fonts_dir, 'ofl')
    apache_dir = os.path.join(args.fonts_dir, 'apache')

    if not os.path.exists(args.output_dir):
        os.mkdir(args.output_dir)

    found_fonts = copy_fonts(ofl_dir, args.lang, args.output_dir,
                             max_fonts=args.num_fonts, found_fonts=0)
    copy_fonts(apache_dir, args.lang, args.output_dir,
               max_fonts=args.num_fonts, found_fonts=found_fonts)


def copy_fonts(src_dir, lang, output_dir, max_fonts, found_fonts=0):
    for font_dir in os.listdir(src_dir):
        if found_fonts >= max_fonts:
            break

        dir_path = os.path.join(src_dir, font_dir)
        ttf_files = [f for f in os.listdir(dir_path) if f.endswith('.ttf')]
        if not ttf_files:
            print(f'No ttf files found', os.listdir(dir_path))
            continue

        metadata_path = os.path.join(dir_path, 'METADATA.pb')
        if not os.path.exists(metadata_path):
            print('no metadata file, directory', font_dir, 'dir path',
                  dir_path, 'dir content', os.listdir(dir_path))
            continue

        with open(metadata_path) as f:
            s = f.read()

        if 'category: "HANDWRITING"' in s and f'subsets: "{lang}"' in s:
            print('found font', font_dir)
            found_fonts += 1

            for name in ttf_files:
                src = os.path.join(dir_path, name)
                dest = os.path.join(output_dir, name)
                shutil.copyfile(src, dest)

    return found_fonts



from lafhterlearn.environment import Environment
from lafhterlearn.evaluation import evaluate
from lafhterlearn.formatters import Formatter
from lafhterlearn.session import SessionDirectoryLayout
from .base import Command


class EvaluateCommand(Command):
    name = 'evaluate'
    help = 'Evaluate a model using a dataset and metrics specified in configuration file'

    def configure_parser(self, parser):
        parser.add_argument('session_dir', type=str,
                            help='Path to the location of session directory')

    def __call__(self, args):
        config = SessionDirectoryLayout(args.session_dir).load_config()
        env = Environment(config)

        metrics = {}
        for task in env.eval_tasks:
            metrics.update(evaluate(task))

        formatter = Formatter()

        final_metrics_string = formatter.format_metrics(metrics)

        whitespaces = ' ' * 150
        print(f'\r{whitespaces}\r{final_metrics_string}')



import importlib
import torch


class Mask:
    def __init__(self, lengths, max_length):
        self.mask = torch.zeros(len(lengths), max_length, dtype=torch.bool)
        self.lengths = lengths

        for i, length in enumerate(lengths):
            self.mask[i, :length] = True

    @property
    def device(self):
        return self.mask.device

    def to(self, device):
        self.mask = self.mask.to(device)
        return self


def add_padding(seq, size, filler):
    seq = list(seq)
    while len(seq) < size:
        seq.append(filler)
    return seq


def pad_sequences(seqs, filler, max_length=None):
    lengths = [len(seq) for seq in seqs]
    max_length = max_length or max(lengths)

    mask = Mask(lengths, max_length)

    padded = [add_padding(seq, max_length, filler) for seq in seqs]
    return padded, mask


def truncate_sequences(seqs, max_length):
    return [s[:max_length] for s in seqs]


def prepare_tf_seqs(transcripts, tokenizer):
    """Prepare a sequence of tokens used for training a decoder with teacher forcing.

    It tokenizes each string, adds a special <start of word> token.

    :param transcripts: strings representing transcripts of word images
    :param tokenizer: tokenizer
    :return: list of integers
    """
    return [tokenizer(t)[:-1] for t in transcripts]


def prepare_targets(transcripts, tokenizer):
    """Converts raw strings into sequences of tokens to be used in loss calculation.

    It tokenizes each string, adds a special <end of word> token.

    :param transcripts: strings representing transcripts of word images
    :param tokenizer: tokenizer
    :return: list of integers
    """
    return [tokenizer(t)[1:] for t in transcripts]


def make_tf_batch(transcripts, tokenizer):
    """Convert raw transcripts into tensor used by decoder network for teacher forcing"""
    transcripts = prepare_tf_seqs(transcripts, tokenizer)
    filler = tokenizer.encode(tokenizer.end)
    padded, mask = pad_sequences(transcripts, filler)
    tf_batch = one_hot_tensor(padded, tokenizer.charset_size)
    return tf_batch, mask


def one_hot_tensor(classes, num_classes):
    """Form a 1-hot tensor from a list of class sequences

    :param classes: list of class sequences (list of lists)
    :param num_classes: total number of available classes
    :return: torch.tensor of shape (batch_size, max_seq_len, num_classes)
    :raise ValueError: if there is any class value that is negative or >= num_classes
    """

    if not hasattr(one_hot_tensor, 'eye'):
        one_hot_tensor.eye = {}

    if num_classes not in one_hot_tensor.eye:
        one_hot_tensor.eye[num_classes] = torch.eye(num_classes, dtype=torch.float32)
    eye = one_hot_tensor.eye[num_classes]
    try:
        tensors = [eye[class_seq] for class_seq in classes]
    except IndexError:
        msg = f'Every class must be a non-negative number less than num_classes={num_classes}. ' \
              f'Got classes {classes}'
        raise ValueError(msg)

    return torch.stack(tensors)


def collate(batch):
    """Split batch into a tuple of lists"""

    num_vars = len(batch[0])
    res = [[] for _ in range(num_vars)]

    for example in batch:
        for i, inp in enumerate(example):
            res[i].append(inp)

    return res


def instantiate_class(dotted_path, *args, **kwargs):
    if '.' not in dotted_path:
        raise ClassImportError(f'Invalid import path: "{dotted_path}"')

    module_path, class_name = split_import_path(dotted_path)
    error_msg = f'Failed to import and instantiate a class "{class_name}" from "{module_path}": '
    try:
        module = importlib.import_module(module_path)
        cls = getattr(module, class_name)
        return cls(*args, **kwargs)
    except Exception as e:
        raise ClassImportError(error_msg + str(e))


def split_import_path(dotted_path):
    idx = dotted_path.rindex('.')
    module_path = dotted_path[:idx]
    name = dotted_path[idx + 1:]
    return module_path, name


class ClassImportError(Exception):
    """Raised when an error occurs during a class importing"""


def full_class_name(obj):
    return obj.__module__ + '.' + obj.__class__.__name__



class CharacterTokenizer:
    start = '<s>'
    end = r'<\s>'
    out_of_charset = '<?>'

    english = "english"

    def __init__(self, charset):
        self.char2index = {}
        self.index2char = {}
        self.charset = charset
        self._build_char_table(charset)

    def __call__(self, value):
        return self.process(value)

    def process(self, text):
        start_token = self.encode(self.start)
        end_token = self.encode(self.end)
        tokens = [self.encode(ch) for ch in text]
        return [start_token] + tokens + [end_token]

    @property
    def charset_size(self):
        return len(self.char2index)

    def _build_char_table(self, charset):
        self._add_char(self.start)
        self._add_char(self.end)
        self._add_char(self.out_of_charset)

        for ch in charset:
            self._add_char(ch)

    def _add_char(self, ch):
        if ch not in self.char2index:
            num_chars = self.charset_size
            self.char2index[ch] = num_chars
            self.index2char[num_chars] = ch

    def encode(self, ch):
        default_char = self.char2index[self.out_of_charset]
        return self.char2index.get(ch, default_char)

    def decode(self, token):
        return self.index2char.get(token, self.out_of_charset)

    def decode_to_string(self, tokens, clean_output=False):
        s = ''.join([self.decode(token) for token in tokens[1:-1]])

        first_char = self.decode(tokens[0])
        last_char = self.decode(tokens[-1])

        if first_char != self.start:
            s = first_char + s

        if last_char != self.end:
            s += last_char

        if clean_output:
            s = s.replace(self.start, '')
            try:
                sentinel_idx = s.index(self.end)
                s = s[:sentinel_idx]
            except ValueError:
                pass

            s = s.replace(self.out_of_charset, '')

        return s



import json
import os

from lafhterlearn.formatters import show_progress_bar
from lafhterlearn.utils import instantiate_class
from .base import Command


class DataCommand(Command):
    name = 'data'
    help = 'Prepare real handwriting data to be used for adaptation stage'

    def configure_parser(self, parser):
        configure_parser(parser)

    def __call__(self, args):
        run(args)


def configure_parser(parser):
    parser.add_argument('data_importer', type=str,
                        help='Fully qualified path (dotted) to Data Importer class')

    parser.add_argument('--kwargs', type=json.loads, default={},
                        help='Optional keyword arguments in JSON format passed to data importer constructor')

    parser.add_argument('--dest', type=str, default='tuning_data',
                        help='Output directory that will store prepared datasets')


def run(args):
    destination_dir = args.dest
    data_importer = instantiate_class(args.data_importer, **args.kwargs)
    create_unlabeled_dataset(data_importer, destination_dir)
    create_labeled_dataset(data_importer, destination_dir)


def create_unlabeled_dataset(data_importer, destination_dir):
    unlabeled_dir = os.path.join(destination_dir, 'unlabeled')
    os.makedirs(unlabeled_dir, exist_ok=True)

    gen = enumerate(data_importer.get_images())
    for i, image in show_progress_bar(gen, desc='Saving images: '):
        image.save(os.path.join(unlabeled_dir, f'{i}.png'))

    spaces = ' ' * 150
    print(f'\r{spaces}')


def create_labeled_dataset(data_importer, destination_dir):
    labeled_dir = os.path.join(destination_dir, 'labeled')

    os.makedirs(labeled_dir, exist_ok=True)

    if data_importer.get_transcribed_images() is not None:
        gen = enumerate(data_importer.get_transcribed_images())
        for i, (image, transcript) in show_progress_bar(gen, desc='Saving images: '):
            image.save(os.path.join(labeled_dir, f'{i}.png'))

            transcripts_path = os.path.join(labeled_dir, 'transcripts.txt')
            with open(transcripts_path, 'a') as f:
                f.write(transcript + '\n')

    spaces = ' ' * 150
    print(f'\r{spaces}')



import os
import re
import random
import PIL
from PIL import Image
from .base import DataImporter


def prepare_iam_dataset(iam_location, max_words=None, train_fraction=0.8, only_letters=False):
    words_dir = os.path.join(iam_location, 'words')

    paths_with_transcripts = parse_examples(iam_location, words_dir, max_words, only_letters)

    return do_random_split(paths_with_transcripts, train_fraction)


def parse_examples(iam_location, words_dir, max_words, only_letters=False):
    transcripts_file = os.path.join(iam_location, 'ascii', 'words.txt')

    lines_skipped = 0
    paths_with_transcripts = []
    with open(transcripts_file) as f:
        for i, line in enumerate(f):
            if max_words and len(paths_with_transcripts) >= max_words:
                break

            try:
                path, gray_level, transcript = parse_line(line, words_dir, only_letters)
                if i + 1 % 10000:
                    print('Words processed:', i + 1)

                paths_with_transcripts.append((path, int(gray_level), transcript))
            except InvalidLineError as e:
                lines_skipped += 1
                print(f'Skipping a problematic line "{line}": {e.args[0]}')

    print(f'total lines processed {i}, lines skipped {lines_skipped}')

    return paths_with_transcripts


def do_random_split(examples, train_fraction):
    indices = list(range(len(examples)))
    random.shuffle(indices)
    train_size = int(len(indices) * train_fraction)

    training_words = [examples[idx] for idx in indices[:train_size]]
    val_words = [examples[idx] for idx in indices[train_size:]]
    return training_words, val_words


def parse_line(line, words_dir, only_letters=False):
    if line.lstrip().startswith('#'):
        raise InvalidLineError(f'Unexpected character "#" at the start of the line: "{line}"')

    parts = re.findall(r'[\w-]+', line)
    image_id = parts[0]
    status = parts[1]
    gray_level = parts[2]

    if status != 'ok':
        raise InvalidLineError(f'Expected status "ok". Got: "{status}"')

    transcript = parts[-1]

    if only_letters and not transcript.isalpha():
        raise InvalidLineError(f'Ignored transcript with non-letter characters: "{transcript}"')

    path = locate_image(words_dir, image_id)
    if not path:
        raise InvalidLineError(f'Image not found in the path: "{path}"')

    try:
        Image.open(path)
    except PIL.UnidentifiedImageError:
        raise InvalidLineError(f'Failure to read an image in the path: "{path}"')

    return path, gray_level, transcript


class InvalidLineError(Exception):
    """Raised when line content does not satisfy certain criteria"""


def locate_image(words_dir, image_id):
    parts = image_id.split('-')
    dir_name = parts[0]
    sub_dir_name = f'{parts[0]}-{parts[1]}'
    file_name = f'{image_id}.png'
    image_path = os.path.join(words_dir, dir_name, sub_dir_name, file_name)
    if os.path.isfile(image_path):
        return image_path


class IAMImporter(DataImporter):
    def __init__(self, iam_location, max_words=None,
                 transcribed_fraction=0.2, only_letters=False):
        self.training_examples, self.val_examples = prepare_iam_dataset(
            iam_location, max_words, 1 - transcribed_fraction, only_letters
        )

    def get_images(self):
        for path, grey_level, _ in self.training_examples:
            yield self.get_cleaned_image(path, grey_level)

    def get_transcribed_images(self):
        for path, grey_level, transcript in self.val_examples:
            image = self.get_cleaned_image(path, grey_level)
            yield image, transcript

    def get_cleaned_image(self, path, grey_level):
        image = Image.open(path)
        return clean_image(image, grey_level)


def clean_image(image, gray_level):
    return image.point(lambda p: 255 if p > gray_level else p)



import os
import json

import yaml

import torch

from lafhterlearn.configuration import Configuration
from lafhterlearn.session import SessionDirectoryLayout, CheckpointKeeper
from lafhterlearn.models import build_networks_spec
from lafhterlearn.environment import create_neural_pipeline
from .base import Command


class CreateSessionCommand(Command):
    name = 'make_session'
    help = 'Create a fresh training session'

    def configure_parser(self, parser):
        configure_parser(parser)

    def __call__(self, args):
        run(args)


def configure_parser(parser):
    parser.add_argument('--config_file', type=str, default='',
                        help='Location of the configuration file (must be a Python module).')


def run(args):
    prepare_session(get_config(args.config_file))


def get_config(config_file) -> Configuration:
    if not config_file:
        return Configuration()

    with open(config_file) as f:
        yml_spec = f.read()

    config_options = yaml.safe_load(yml_spec)

    config = Configuration()

    for name, value in config_options.items():
        if name == 'device':
            if value != 'auto':
                config.device = torch.device(value)
        else:
            setattr(config, name, value)

    return config


def prepare_session(config):
    if os.path.exists(config.session_dir):
        print(f'Session already exists in "{config.session_dir}"')
        return

    session_layout = SessionDirectoryLayout(config.session_dir)
    session_layout.make_session_dir()
    session_layout.make_checkpoints_dir()

    spec = build_networks_spec(charset=config.charset,
                               image_height=config.image_height,
                               hidden_size=config.hidden_size,
                               **config.decoder_params)

    device = torch.device(config.device)
    neural_pipeline = create_neural_pipeline(device, spec, config)

    keeper = CheckpointKeeper(session_layout.checkpoints)

    with open(session_layout.model_spec, 'w') as f:
        f.write(json.dumps(spec))

    neural_pipeline.encoder.to(neural_pipeline.device)
    neural_pipeline.decoder.to(neural_pipeline.device)
    keeper.make_new_checkpoint(neural_pipeline, device, 0, metrics={})

    json_str = config.to_json()

    config_save_path = os.path.join(config.session_dir, "config.json")
    with open(config_save_path, 'w') as f:
        f.write(json_str)



from collections import OrderedDict

import h5py
from nltk.lm.preprocessing import pad_both_ends
from nltk.util import ngrams
import numpy as np


def load_sparse_array(indices, values, size):
    a = SparseArray(size)
    for i, v in zip(indices, values):
        a[i] = v
    return a


class SparseArray:
    def __init__(self, size, fill=0):
        self.size = size
        self.fill = fill
        self.data = OrderedDict()

    @property
    def indices(self):
        return list(self.data.keys())

    @property
    def values(self):
        return list(self.data.values())

    def __len__(self):
        return self.size

    def __getitem__(self, idx):
        if 0 <= idx < len(self):
            return self.data.get(idx, self.fill)

        raise IndexError(f'index out of bounds: {idx}')

    def __setitem__(self, key, value):
        if 0 <= key < len(self):
            self.data[key] = value
        else:
            raise IndexError(f'index out of bounds: {key}')


class SparseMatrix:
    """Resizable ragged matrix with sparse rows"""
    def __init__(self, h5obj, group_name, cols):
        if group_name not in h5obj:
            group = h5obj.create_group(group_name)
        else:
            group = h5obj[group_name]

        self.group = group
        self.cols = cols

        indices_name = "indices"
        values_name = "values"

        self.indices = RaggedMatrix(group, indices_name)
        self.values = RaggedMatrix(group, values_name)

    def add_empty_row(self):
        self.indices.append_row([])
        self.values.append_row([])

    def bulk_update(self, row, indices, values):
        """Replace/put values in specified positions in a given row"""
        sparse_array = self[row]
        for idx, value in zip(indices, values):
            sparse_array[idx] = value

        self.indices.update_row(row, sparse_array.indices)
        self.values.update_row(row, sparse_array.values)

    def __getitem__(self, idx):
        indices_row = self.indices[idx]
        values_row = self.values[idx]
        return load_sparse_array(indices_row, values_row, self.cols)

    def __len__(self):
        return len(self.indices)


class ProbabilityDistribution:
    def __init__(self, sparse_counts_array):
        self.sparse_counts_array = sparse_counts_array
        self.rng = np.random.default_rng()

    def sample(self):
        counts = self.sparse_counts_array.values
        pmf = counts / sum(counts)

        indices = list(range(counts))
        return self.rng.choice(indices, p=pmf)


class BaseH5Matrix:
    def __init__(self, h5obj, ds_name, create_dataset):
        if ds_name not in h5obj:
            dset = create_dataset()
        else:
            dset = h5obj[ds_name]

        self.dset = dset

    def append_row(self, arr):
        current_row = len(self)
        self.dset.resize(current_row + 1, axis=0)
        self.dset[current_row] = arr[:]

    def __getitem__(self, idx):
        if isinstance(idx, int):
            if 0 <= idx < len(self):
                return self.dset[idx]
            raise IndexError()

        raise IndexError('index should be an integer, not a tuple')

    def __len__(self):
        return len(self.dset)


class ExpandableMatrix(BaseH5Matrix):
    """Matrix backed by H5 with fixed number of columns, but dynamic number of rows"""
    def __init__(self, h5obj, ds_name, cols):
        def create_dataset():
            return h5obj.create_dataset(ds_name, (0, cols), maxshape=(None, cols), dtype='i')
        super().__init__(h5obj, ds_name, create_dataset)
        self.cols = cols

    def append_row(self, arr):
        if len(arr) != self.cols:
            raise ValueError(f'expects rows of length {self.cols}, got {len(arr)}')
        super().append_row(arr)


class RaggedMatrix(BaseH5Matrix):
    """Resizable matrix with varying number of columns"""
    def __init__(self, h5obj, ds_name):
        def create_dataset():
            dt = h5py.vlen_dtype(np.dtype('int32'))
            return h5obj.create_dataset(ds_name, (0,), maxshape=(None,), dtype=dt)
        super().__init__(h5obj, ds_name, create_dataset)

    def update_row(self, row, values):
        self.dset[row] = values[:]


def build_vocab(words, unk_cutoff=10):
    from collections import Counter
    counter = Counter(words)
    vocab = set(word for word, count in counter.items() if count >= unk_cutoff)
    vocab.add('<UNK>')
    vocab.add('<s>')
    vocab.add(r'<\s>')
    return vocab


def get_word_stream(get_corpus, n):
    for sent in get_corpus():
        yield from pad_both_ends(sent, n=n)


def get_ngram_stream(get_corpus, n):
    return ngrams(get_word_stream(get_corpus, n), n=n)



import random
import csv
import string

import numpy as np
import h5py

from .ngrams import NgramModel


class UniformSampler:
    def __init__(self, words):
        self.words = words

    def __call__(self):
        return random.choice(self.words)

    @classmethod
    def from_file(cls, path):
        with open(path) as f:
            words = [word.strip() for word in f if word.strip()]
            return cls(words)


class FrequencyBasedSampler:
    def __init__(self, words, frequencies, sampling_batch_size=1024):
        self.words = words
        self.frequencies = frequencies
        self.rng = np.random.default_rng()

        self.batch_size = sampling_batch_size
        self.buffer = []

    def __call__(self):
        # potential race condition when using data loaders with num_workers > 0
        if not self.buffer:
            self.buffer = list(self.rng.choice(self.words, size=self.batch_size,
                                               p=self.frequencies))

        return self.buffer.pop()

    @classmethod
    def from_file(cls, path):
        with open(path, newline='') as csvfile:
            reader = csv.reader(csvfile, delimiter=',', quotechar='|')

            words = []
            freqs = []
            for word, freq in reader:
                words.append(word)
                freqs.append(freq)

            return cls(words, freqs)


class NgramBasedSampler:
    def __init__(self, ngram_model, num_words=10):
        self.model = ngram_model
        self.num_words = num_words

    def __call__(self):
        # todo: make it faster
        words = self.model.generate(num_words=self.num_words)
        prev = words[0]

        seq = [prev]
        for word in words[1:]:
            if ((prev == "'" and word in ["s", "m", "t", "ve"]) or
                    (prev in string.punctuation and word in string.punctuation) or
                    (prev.isalnum() and word in string.punctuation)):
                seq.append(word)
            else:
                seq.append(' ')
                seq.append(word)
            prev = word
        return ''.join(seq)

    @classmethod
    def from_file(cls, path):
        f = h5py.File(path)
        model = NgramModel.from_h5file(f)
        return cls(model)



import logging
from dataclasses import dataclass
from typing import Any

import torch
from lafhterlearn.utils import make_tf_batch
from .tokenizer import CharacterTokenizer
from .image_pipelines import ImagePipeline
from .models import ImageEncoder, AttendingDecoder


logger = logging.getLogger(__name__)


@dataclass
class EncoderDecoder:
    encoder: ImageEncoder
    decoder: AttendingDecoder
    device: torch.device

    def debug_attention(self, image_batch):
        image_batch = image_batch.to(self.device)
        encodings = self.encoder(image_batch)
        return self.decoder.debug_attention(encodings)

    def predict(self, image_batch, transcripts=None):
        image_batch = image_batch.to(self.device)
        if transcripts is not None:
            transcripts = transcripts.to(self.device)

        encodings = self.encoder(image_batch)
        return self.decoder(encodings, transcripts)

    def __call__(self, image_batch, transcripts=None):
        return self.predict(image_batch, transcripts)


@dataclass
class TrainableEncoderDecoder(EncoderDecoder):
    encoder_optimizer: Any
    decoder_optimizer: Any

    def zero_grad(self):
        self.encoder.zero_grad()
        self.decoder.zero_grad()

    def step(self):
        self.encoder_optimizer.step()
        self.decoder_optimizer.step()

    def train_mode(self):
        self.encoder.train()
        self.decoder.train()

    def eval_mode(self):
        self.encoder.eval()
        self.decoder.eval()


@dataclass
class WordRecognitionPipeline:
    neural_pipeline: TrainableEncoderDecoder
    tokenizer: CharacterTokenizer
    image_preprocessor: ImagePipeline
    show_attention: bool = False

    def __call__(self, images, transcripts=None):
        """Given a list of PIL images and (optionally) corresponding list of text transcripts"""
        return self._try_recognize(images, transcripts)

    def _try_recognize(self, images, transcripts):
        try:
            return self._do_recognize(images, transcripts)
        except torch.cuda.OutOfMemoryError:
            widths_with_transcripts = [(im.width, t)
                                       for im, t in zip(images, transcripts)]
            largest_pair = max(widths_with_transcripts, key=lambda pair: pair[0])

            logger.exception('Recognition failed: out of memory.'
                             'Longest image width {}. '
                             'image transcript "{}"'.format(*largest_pair))
            raise

    def _do_recognize(self, images, transcripts):
        image_batch = self.image_preprocessor(images)

        if transcripts is not None:
            transcripts, _ = make_tf_batch(transcripts, self.tokenizer)

        if self.show_attention:
            return self.neural_pipeline.debug_attention(image_batch)
        else:
            return self.neural_pipeline(image_batch, transcripts)
